"""
Model Generator for OpenSees ~ tributary area analysis
"""


#                          __
#   ____  ____ ___  ____ _/ /
#  / __ \/ __ `__ \/ __ `/ /
# / /_/ / / / / / / /_/ /_/
# \____/_/ /_/ /_/\__, (_)
#                /____/
#
# https://github.com/ioannis-vm/OpenSees_Model_Generator

from __future__ import annotations
from typing import TYPE_CHECKING
from typing import no_type_check
from dataclasses import dataclass, field
import numpy as np
import numpy.typing as npt
from .. import mesh
from ..ops.node import Node
from .. import common
from ..ops.element import ElasticBeamColumn
from ..ops.element import DispBeamColumn
if TYPE_CHECKING:
    from ..load_case import LoadCase
    from ..level import Level
    from ..mesh import Edge, Vertex, Halfedge

# pylint: disable=no-member
# pylint: disable=import-outside-toplevel

nparr = npt.NDArray[np.float64]
PolygonShape = list[tuple[float, float]]


@dataclass
class PolygonLoad:
    """
    Polygonal loaded area
    """
    name: str
    value: float
    outside_shape: PolygonShape
    holes: list[PolygonShape]
    massless: bool


@dataclass(repr=False)
class TributaryAreaAnalysisData:
    """
    Data generated by the tributary area analysis code
    """
    # collects the defined edges
    edges: dict[int, Edge] = field(default_factory=dict)
    # collects the defined vertices
    vertices: dict[int, Vertex] = field(default_factory=dict)
    # maps an edge.uid to a node or line element
    edge_map: dict[int, Node | ElasticBeamColumn | DispBeamColumn] = \
        field(default_factory=dict)
    # maps an edge.uid to its corresponding tributary area
    edge_area: dict[int, float] = field(default_factory=dict)
    # maps an edge.uid to its corresponding tributary polygon
    edge_polygons: dict[int, PolygonShape] = field(default_factory=dict)
    # maps a node.uid to a vertex
    vertex_map: dict[int, Vertex] = field(default_factory=dict)
    # used to map zerolength element-induced coinciding nodes
    zn_map: dict[int, int] = field(default_factory=dict)
    # used to map panel zone nodes to their primary node
    pz_node: dict[int, dict[str, Node]] = field(default_factory=dict)


@dataclass
class TributaryAreaAnaysis:
    """
    Tributary area analysis object
    """
    parent_loadcase: LoadCase
    parent_level: Level
    polygon_loads: list[PolygonLoad] = field(
        default_factory=list)
    data: TributaryAreaAnalysisData = field(
        default_factory=TributaryAreaAnalysisData)

    @no_type_check
    def run(self, load_factor=1.00, massless_load_factor=1.00):
        """
        Performs tributary area analysis
        """

        try:
            import skgeom as sg
        except ModuleNotFoundError:
            msg = "One day, a custom implementation of the"
            msg += "straight skeleton algorithm might be added to osmg.\n"
            msg += "Until that day, the scikit-geometry package is required.\n"
            msg += "Please install scikit-geometry.\n"
            msg += "https://github.com/scikit-geometry/scikit-geometry"
            print()

        lvl = self.parent_level
        all_components = list(lvl.components.values())
        horizontal_elements = []
        panel_zones = []
        for component in all_components:
            if component.component_purpose == 'horizontal_component':
                horizontal_elements.append(component)
            elif component.component_purpose == 'steel_W_panel_zone':
                panel_zones.append(component)

        # # plotting - used while developing the code
        # subset_model = mdl.initialize_empty_copy('subset_1')
        # for component in horizontal_elements+panel_zones:
        #     mdl.transfer_component(subset_model, component)
        # show(subset_model)

        # define vertices, edges, mapping
        # we "flatten" the components in the xy plane.
        # for joints, we only consider the top-most line elements
        # and we treat them the same as rigid offsets.

        edges = self.data.edges
        vertices = self.data.vertices
        edge_map = self.data.edge_map
        edge_area = self.data.edge_area
        edge_polygons = self.data.edge_polygons
        vertex_map = self.data.vertex_map
        zn_map = self.data.zn_map
        pz_node = self.data.pz_node
        # clear any previous reults
        edges = {}
        vertices = {}
        edge_map = {}
        edge_area = {}
        edge_polygons = {}
        vertex_map = {}
        zn_map = {}
        pz_node = {}

        for panel_zone in panel_zones:
            back_nd = panel_zone.external_nodes.named_contents['middle_back']
            front_nd = panel_zone.external_nodes.named_contents['middle_front']
            main_nd = panel_zone.external_nodes.named_contents['top_node']
            back_offset: nparr = (np.array(main_nd.coords)
                                  - np.array(back_nd.coords))
            front_offset: nparr = (np.array(main_nd.coords)
                                   - np.array(front_nd.coords))
            pz_node[back_nd.uid] = {
                'substitute_node': main_nd,
                'additional_offset': -back_offset[0:2]
            }
            pz_node[front_nd.uid] = {
                'substitute_node': main_nd,
                'additional_offset': -front_offset[0:2]
            }

        for comp in horizontal_elements:

            line_elements = (list(comp.elastic_beamcolumn_elements.values())
                             + list(comp.disp_beamcolumn_elements
                                    .values()))  # type: ignore

            zerolength_elements = list(comp.zerolength_elements.values())
            for zelm in zerolength_elements:
                zn_map[zelm.nodes[1].uid] = zelm.nodes[0].uid

            for elm in line_elements:
                n_i = elm.nodes[0]
                eo_i = elm.geomtransf.offset_i[0:2]
                if n_i.uid in zn_map:
                    n_i = comp.internal_nodes[zn_map[n_i.uid]]
                n_j = elm.nodes[1]
                eo_j = elm.geomtransf.offset_j[0:2]
                if n_j.uid in zn_map:
                    n_j = comp.internal_nodes[zn_map[n_j.uid]]
                if n_i.uid in pz_node:
                    sub_data = pz_node[n_i.uid]
                    n_i = sub_data['substitute_node']
                    eo_i = eo_i.copy() + sub_data['additional_offset']
                if n_j.uid in pz_node:
                    sub_data = pz_node[n_j.uid]
                    n_j = sub_data['substitute_node']
                    eo_j = eo_j.copy() + sub_data['additional_offset']

                if n_i.uid not in vertex_map:
                    vrt_i = mesh.Vertex(
                        (n_i.coords[0],
                         n_i.coords[1]))
                    vertex_map[n_i.uid] = vrt_i
                    vertices[vrt_i.uid] = vrt_i
                else:
                    vrt_i = vertex_map[n_i.uid]
                if n_j.uid not in vertex_map:
                    vrt_j = mesh.Vertex(
                        (n_j.coords[0],
                         n_j.coords[1]))
                    vertex_map[n_j.uid] = vrt_j
                    vertices[vrt_j.uid] = vrt_j
                else:
                    vrt_j = vertex_map[n_j.uid]

                if np.linalg.norm(eo_i) >= common.EPSILON:
                    # there is a rigid offset and/or panel zone
                    point = np.array(vrt_i.coords) + eo_i
                    vrt_oi = mesh.Vertex((point[0], point[1]))
                    vertices[vrt_oi.uid] = vrt_oi
                    edg_oi = mesh.Edge(vrt_i, vrt_oi)
                    edges[edg_oi.uid] = edg_oi
                    edge_map[edg_oi.uid] = n_i
                    connecting_vertex_i = vrt_oi
                else:
                    connecting_vertex_i = vrt_i
                if np.linalg.norm(eo_j) >= common.EPSILON:
                    # there is a rigid offset and/or panel zone
                    point = np.array(vrt_j.coords) + eo_j
                    vrt_oj = mesh.Vertex(
                        (point[0], point[1]))
                    vertices[vrt_oj.uid] = vrt_oj
                    edg_oj = mesh.Edge(vrt_j, vrt_oj)
                    edges[edg_oj.uid] = edg_oj
                    edge_map[edg_oj.uid] = n_j
                    connecting_vertex_j = vrt_oj
                else:
                    connecting_vertex_j = vrt_j
                edg_interior = mesh.Edge(connecting_vertex_i,
                                         connecting_vertex_j)
                edge_map[edg_interior.uid] = elm
                edges[edg_interior.uid] = edg_interior

        # # plotting - used while developing the code
        # import pandas as pd
        # edf = pd.DataFrame(np.zeros((len(edges)*3, 2)))
        # edf.columns = ['x', 'y']
        # enames = []
        # for i, edge_key in enumerate(edges):
        #     edge = edges[edge_key]
        #     edf.loc[i*3+0, 'x':'y'] = edge.v_i.coords
        #     edf.loc[i*3+1, 'x':'y'] = edge.v_j.coords
        #     edf.loc[i*3+2, 'x':'y'] = (None, None)
        #     enames.extend((f'E{edge.uid} to V{edge.v_i.uid}',
        #                    f'E{edge.uid} to V{edge.v_j.uid}', None))
        # vdf = pd.DataFrame(np.zeros((len(vertices), 2)))
        # vdf.columns = ['x', 'y']
        # vnames = []
        # for i, vertex_key in enumerate(vertices):
        #     vertex = vertices[vertex_key]
        #     vdf.loc[i, 'x':'y'] = vertex.coords
        #     vnames.append(vertex.uid)
        # edf = edf + np.random.normal(0.00, 0.10, edf.shape)
        # import plotly.express as px
        # import plotly.graph_objects as go
        # fig1 = px.line(edf, x='x', y='y', hover_name=enames)
        # fig2 = px.scatter(vdf, x='x', y='y', hover_name=vnames,
        #                   color=['red']*len(vnames))
        # fig = go.Figure(data=fig1.data + fig2.data)
        # fig.show()

        halfedges = mesh.define_halfedges(list(edges.values()))
        loops = mesh.obtain_closed_loops(halfedges)
        external, internal, trivial = \
            mesh.orient_loops(loops)
        # Sanity checks.
        mesh.sanity_checks(external, trivial)

        for internal_loop in internal:
            poly = sg.Polygon([h.vertex.coords for h in internal_loop])
            skel = sg.skeleton.create_interior_straight_skeleton(poly)
            # todo: what we need to get rid of this is to end up with a list
            # of halfedges defining the subloops.

            # something like:
            # subloops = mesh.bisector_subdivision(internal_loop)

            subloops: list[list[Halfedge]] = []

            def is_in_some_subloop(halfedge, loops):
                for loop in loops:
                    for other_halfedge in loop:
                        if (other_halfedge.vertex.point ==
                            halfedge.vertex.point and
                                other_halfedge.next.vertex.point ==
                                halfedge.next.vertex.point):
                            return True
                return False

            for halfedge in skel.halfedges:
                if subloops:
                    if is_in_some_subloop(halfedge, subloops):
                        continue
                subloop = [halfedge]
                nxt = halfedge.next
                while nxt.vertex.point != halfedge.vertex.point:
                    subloop.append(nxt)
                    nxt = nxt.next
                subloops.append(subloop)

            subloop_areas = [float(sg.Polygon(
                [h.vertex.point for h in subloop]).area())
                for subloop in subloops]
            outer = min(subloop_areas)  # Remove the exterior loop
            index = subloop_areas.index(outer)
            del subloops[index]
            del subloop_areas[index]

            for i, subloop in enumerate(subloops):
                area = subloop_areas[i]
                loop_edges = [h.edge for h in internal_loop]
                for halfedge in subloop:
                    for edge in loop_edges:
                        v_i = sg.Point2(*edge.v_i.coords)
                        v_j = sg.Point2(*edge.v_j.coords)
                        pt_1 = halfedge.vertex.point
                        pt_2 = halfedge.next.vertex.point
                        if ((pt_1 == v_i and pt_2 == v_j) or
                                (pt_1 == v_j and pt_2 == v_i)):
                            if edge.uid in edge_area:
                                edge_area[edge.uid] += area
                            else:
                                edge_area[edge.uid] = area
                            if edge.uid in edge_polygons:
                                edge_polygons[edge.uid].append(
                                    [(float(h.vertex.point.x()),
                                      float(h.vertex.point.y()))
                                     for h in subloop])
                            else:
                                edge_polygons[edge.uid] = [
                                    [(float(h.vertex.point.x()),
                                      float(h.vertex.point.y()))
                                     for h in subloop]]

        # # plotting - used while developing the code
        # import pandas as pd
        # import plotly.express as px
        # import plotly.graph_objects as go

        # x_vals = []
        # y_vals = []
        # colors = []
        # for edge in edges.values():
        #     x_vals.extend([edge.v_i.coords[0], edge.v_j.coords[0], None])
        #     y_vals.extend([edge.v_i.coords[1], edge.v_j.coords[1], None])
        #     colors.extend(['black', 'black', None])

        # fig1 = go.Figure(go.Scatter(x=x_vals, y=y_vals,
        #                             mode='lines',
        #                             line_color='black',
        #                             name='Elements'))

        # x_vals = []
        # y_vals = []
        # for edge in edges.values():
        #     for poly in edge_polygons[edge.uid]:
        #         poly_rot = poly[1:]+poly[:1]
        #         for vi, vj in zip(poly, poly_rot):
        #             x_vals.extend((vi[0], vj[0], None))
        #             y_vals.extend((vi[1], vj[1], None))

        # fig2 = go.Figure(go.Scatter(x=x_vals, y=y_vals,
        #                             mode='lines',
        #                             line_color='blue',
        #                             name='Regions'))

        # fig = go.Figure(data=fig2.data + fig1.data)
        # fig.update_yaxes(
        #     scaleanchor = "x",
        #     scaleratio = 1,
        #   )
        # fig.show()

        # apply loads and mass
        # todo (future): account for the shape of the loaded area as well
        mdl = self.parent_level.parent_model
        if mdl.settings.imperial_units:
            g_const = common.G_CONST_IMPERIAL
        else:
            g_const = common.G_CONST_SI
        lcase = self.parent_loadcase
        for uid, edge in edges.items():
            area = edge_area[uid]
            loaded_elm = edge_map[uid]
            for load in self.polygon_loads:
                if load.massless:
                    cmult = massless_load_factor
                else:
                    cmult = load_factor
                surf_load = load.value
                load_val = surf_load * area
                if isinstance(loaded_elm, Node):
                    lcase.node_loads[
                        loaded_elm.uid
                    ].add(
                        np.array([0.00, 0.00,
                                  -load_val * cmult,
                                  0.00, 0.00, 0.00])
                    )
                    if not load.massless:
                        lcase.node_mass[
                            loaded_elm.uid
                        ].add(
                            np.array([load_val/g_const,
                                      load_val/g_const,
                                      load_val/g_const,
                                      0.00, 0.00, 0.00])
                        )

                elif isinstance(loaded_elm,
                                (ElasticBeamColumn,
                                 DispBeamColumn)):
                    length = loaded_elm.clear_length()
                    load_per_length = load_val / length
                    udl_obj = lcase.line_element_udl[
                        loaded_elm.uid
                    ]
                    udl_obj.add_glob(np.array(
                        [0.00, 0.00, -load_per_length * cmult]))
                    if not load.massless:
                        half_mass = (load_val/2.00/g_const)
                        lcase.node_mass[
                            loaded_elm.nodes[0].uid].add(
                                np.array([half_mass]*3+[0.00]*3))
                        lcase.node_mass[
                            loaded_elm.nodes[1].uid].add(
                                np.array([half_mass]*3+[0.00]*3))

                else:
                    raise TypeError('This should never happen!')



"""
Performs coordinate transformations
"""

#                          __
#   ____  ____ ___  ____ _/ /
#  / __ \/ __ `__ \/ __ `/ /
# / /_/ / / / / / / /_/ /_/
# \____/_/ /_/ /_/\__, (_)
#                /____/
#
# https://github.com/ioannis-vm/OpenSees_Model_Generator

import numpy as np
import numpy.typing as npt
from . import common

nparr = npt.NDArray[np.float64]


def rotation_matrix_2d(ang: float) -> nparr:
    """
    Return a 2D transformation matrix
    """
    return np.array([[np.cos(ang), -np.sin(ang)], [np.sin(ang), np.cos(ang)]])


def rotation_matrix_3d(
        axis: nparr,
        theta: float) -> nparr:
    """
    Return the rotation matrix associated with counterclockwise
    rotation about the given axis by theta radians.
    Courtesy of
    https://stackoverflow.com/questions/6802577/rotation-of-3d-vector
    """
    v_a = np.cos(theta / 2.0)
    v_b, v_c, v_d = -axis * np.sin(theta / 2.0)
    v_aa, v_bb, v_cc, v_dd = v_a * v_a, v_b * v_b, v_c * v_c, v_d * v_d
    v_bc, v_ad, v_ac, v_ab, v_bd, v_cd = (v_b * v_c, v_a * v_d, v_a * v_c, v_a
                                          * v_b, v_b * v_d, v_c * v_d)
    return np.array([[v_aa + v_bb - v_cc - v_dd, 2
                      * (v_bc + v_ad), 2 * (v_bd - v_ac)],
                     [2 * (v_bc - v_ad), v_aa + v_cc
                      - v_bb - v_dd, 2 * (v_cd + v_ab)],
                     [2 * (v_bd + v_ac), 2 * (v_cd - v_ab),
                      v_aa + v_dd - v_bb - v_cc]])


def transformation_matrix(vec_x: nparr,
                          vec_y: nparr,
                          vec_z: nparr
                          ) -> nparr:
    """
    Returns a transformation matrix that transforms points from
    the coordinate system in which the x, y and z axes are expressed,
    to the local coordinate system defined by them.
    Args:
        vec_x (nparr): Local x axis expressed in the global system
        vec_y (nparr): (similar)
        vec_z (nparr): (similar)
    Returns:
        (nparr): global to local transformation matrix.
    Note: For orthogonal axes, transpose to obtain the inverse transform.
    """
    tr_global_to_local: nparr = \
        np.vstack((vec_x, vec_y, vec_z))
    return tr_global_to_local


def local_axes_from_points_and_angle(point_i: nparr,
                                     point_j: nparr,
                                     ang: float) -> \
        tuple[nparr,
              nparr,
              nparr]:
    """
    Given a start point, and end point, and an angle,
    obtain the local coordinate system of a linear element.
    Args:
        point_i (nparr): Start point
        point_j (nparr): End point
        and (float): Parameter that controls the rotation of the
            section around the x-axis. Counterclockwise rotation is
            posotive. 0.00 corresponds to:
              vertical elements whose local z axis coincides with
                                the local x axis
              horizontal elements whose local z axis is horizontal.
    """
    # x
    x_axis = point_j - point_i
    x_axis = x_axis / np.linalg.norm(x_axis)
    # y and z
    diff = np.abs(
        np.linalg.norm(x_axis - np.array([0.00, 0.00, -1.00])))
    if diff < common.EPSILON:
        # vertical case
        z_axis: nparr = np.array([np.cos(ang), np.sin(ang), 0.0])
        y_axis: nparr = np.cross(z_axis, x_axis)
    else:
        # note: all vertical elements must have node i higher than node j
        # we raise an error otherwise.
        # check:
        diff = np.abs(
            np.linalg.norm(x_axis - np.array([0.00, 0.00, 1.00])))
        if diff < common.EPSILON:
            raise ValueError("Vertical element defined upside down")
        # not vertical case
        up_direction: nparr = np.array([0.0, 0.0, 1.0])
        # orthogonalize with respect to x_vec
        y_axis = \
            up_direction - np.dot(up_direction, x_axis) * x_axis
        # ..and normalize
        y_axis = y_axis / np.linalg.norm(y_axis)
        y_axis = np.dot(rotation_matrix_3d(
            x_axis, ang), y_axis)
        # determine z axis from the cross-product
        z_axis = np.cross(x_axis, y_axis)

    return x_axis, y_axis, z_axis  # type: ignore


def offset_transformation(offset: nparr,
                          u_vec: nparr,
                          r_vec: nparr) -> nparr:
    """
    Obtain the displacement of the end of a rigid offeset
    by specifying the displacement and rotation of the
    other end.
    Args:
        offset: Vector pointing from the node of the rigid
                offset of which the displacement is known
                to the node where we want to obtain the
                displacement.
    Returns:
        Displacements at the other end.
    """
    t_rigid: nparr = np.array([[0.00, +offset[2], -offset[1]],
                               [-offset[2], 0.00, +offset[0]],
                               [+offset[1], -offset[0], 0.00]])
    return u_vec + t_rigid @ r_vec



"""
Model Generator for OpenSees ~ section generator
"""

#                          __
#   ____  ____ ___  ____ _/ /
#  / __ \/ __ `__ \/ __ `/ /
# / /_/ / / / / / / /_/ /_/
# \____/_/ /_/ /_/\__, (_)
#                /____/
#
# https://github.com/ioannis-vm/OpenSees_Model_Generator

from __future__ import annotations
from typing import TYPE_CHECKING
from typing import Type
from dataclasses import dataclass
import json
import pkgutil
import numpy as np
import numpy.typing as npt
from ..ops.section import SectionComponent
from ..ops.section import ElasticSection
from ..ops.section import FiberSection
from ..gen import mesh_shapes
from .mesh_shapes import rect_mesh
if TYPE_CHECKING:
    from ..model import Model
    from ..ops.section import Section

nparr = npt.NDArray[np.float64]


@dataclass(repr=False)
class SectionGenerator:
    """
    Used to populate the section repository of a model.
    """
    model: Model

    def generate_generic_elastic(
            self,
            name: str,
            e_times_a: float,
            e_times_i: float,
            g_times_j: float
    ):
        """
        Generates a generic elastic section with the specified
        properties
        """
        sec = ElasticSection(
            name=name,
            uid=self.model.uid_generator.new('section'),
            outside_shape=None,
            snap_points=None,
            e_mod=1.00,
            area=e_times_a,
            i_y=e_times_i,
            i_x=e_times_i,
            g_mod=1.00,
            j_mod=g_times_j,
            sec_w=0.00,
        )
        if self.model.settings.imperial_units:
            y_max = +10.00
            y_min = -10.00
            z_max = +6.00
            z_min = -6.00
            sec.outside_shape = rect_mesh(12.0, 20.0)
        else:
            y_max = +0.25
            y_min = -0.25
            z_max = +0.15
            z_min = -0.15
            sec.outside_shape = rect_mesh(0.30, 0.50)
        snap_points: dict[str, nparr] = {
            'centroid': np.array([0., 0.]),
            'top_center': np.array([0., -y_max]),
            'top_left': np.array([-z_min, -y_max]),
            'top_right': np.array([-z_max, -y_max]),
            'center_left': np.array([-z_min, 0.]),
            'center_right': np.array([-z_max, 0.]),
            'bottom_center': np.array([0., -y_min]),
            'bottom_left': np.array([-z_min, -y_min]),
            'bottom_right': np.array([-z_max, -y_min])
        }
        sec.snap_points = snap_points
        self.model.elastic_sections.add(sec)
        return sec

    def load_aisc_from_database(
            self,
            sec_shape_designation: str, labels: list[str],
            ops_material: str, physical_material: str,
            sec_type: Type[Section]):
        """
        Loads a section from the AISC steel section database.
        """
        ops_mat = self.model.uniaxial_materials.retrieve_by_attr(
            'name', ops_material)
        phs_mat = self.model.physical_materials.retrieve_by_attr(
            'name', physical_material)
        filename = '../../../section_data/sections.json'
        contents = pkgutil.get_data(__name__, filename)
        assert isinstance(contents, bytes)
        section_dictionary = json.loads(contents)
        assert self.model.settings.imperial_units, 'SI not supported'
        for label in labels:
            try:
                sec_data = section_dictionary[label]
            except KeyError:
                raise KeyError(f'Section {label} not found in file.')
            if sec_shape_designation == 'W':
                sec_b = sec_data['bf']
                sec_h = sec_data['d']
                sec_tw = sec_data['tw']
                sec_tf = sec_data['tf']
                area = sec_data['A']
                outside_shape = mesh_shapes.w_mesh(
                    sec_b, sec_h, sec_tw, sec_tf, area)
                bbox = outside_shape.bounding_box()
                z_min, y_min, z_max, y_max = bbox.flatten()
                snap_points: dict[str, nparr] = {
                    'centroid': np.array([0., 0.]),
                    'top_center': np.array([0., -y_max]),
                    'top_left': np.array([-z_min, -y_max]),
                    'top_right': np.array([-z_max, -y_max]),
                    'center_left': np.array([-z_min, 0.]),
                    'center_right': np.array([-z_max, 0.]),
                    'bottom_center': np.array([0., -y_min]),
                    'bottom_left': np.array([-z_min, -y_min]),
                    'bottom_right': np.array([-z_max, -y_min])
                }
                if sec_type.__name__ == 'FiberSection':
                    main_part = SectionComponent(
                        outside_shape,
                        {},
                        ops_mat,
                        phs_mat)
                    sec_fib = FiberSection(
                        label,
                        self.model.uid_generator.new('section'),
                        outside_shape,
                        {'main': main_part},
                        sec_data['J'],
                        snap_points)
                    self.model.fiber_sections.add(sec_fib)
                elif sec_type.__name__ == 'ElasticSection':
                    sec_el = ElasticSection(
                        label,
                        self.model.uid_generator.new('section'),
                        phs_mat.e_mod,
                        sec_data['A'],
                        sec_data['Iy'],
                        sec_data['Ix'],
                        phs_mat.g_mod,
                        sec_data['J'],
                        sec_data['W'] / 12.00,  # lb/in
                        outside_shape,
                        snap_points,
                        properties=sec_data)
                    self.model.elastic_sections.add(sec_el)
                else:
                    raise ValueError(
                        f'Unsupported section type: {sec_type.__name__}')
            else:
                raise ValueError(
                    'Unsupported section designtation:'
                    f' {sec_shape_designation}')



"""
Model Generator for OpenSees ~ model
"""

#                          __
#   ____  ____ ___  ____ _/ /
#  / __ \/ __ `__ \/ __ `/ /
# / /_/ / / / / / / /_/ /_/
# \____/_/ /_/ /_/\__, (_)
#                /____/
#
# https://github.com/ioannis-vm/OpenSees_Model_Generator

# pylint: disable=W1512

from __future__ import annotations
from typing import TYPE_CHECKING
from dataclasses import dataclass, field
import numpy as np
import numpy.typing as npt
from shapely.geometry import Polygon as shapely_Polygon   # type: ignore
from shapely.geometry import Point   # type: ignore

from .gen.uid_gen import UIDGenerator
from . import collections
from .level import Level

if TYPE_CHECKING:
    from .ops.node import Node
    from .ops.element import ElasticBeamColumn
    from .ops.element import DispBeamColumn
    from .ops.element import ZeroLength
    from .component_assembly import ComponentAssembly
    from .ops.section import ElasticSection
    from .ops.section import FiberSection
    from .ops.uniaxial_material import UniaxialMaterial
    from .physical_material import PhysicalMaterial

nparr = npt.NDArray[np.float64]


def transfer_component(
        other: Model,
        component: ComponentAssembly):
    """
    Transfers a single component assembly from one model to
    another, assuming the other model was generated with the
    `initialize_empty_copy` method.
    """
    # note: we don't copy the component assemblies and their contents.
    # we just add the same objects to the other model.
    level = component.parent_collection.parent
    other_level = other.levels.retrieve_by_attr('uid', level.uid)
    for node in component.external_nodes.values():
        if node.uid not in other_level.nodes:
            other_level.nodes.add(node)
    other_level.components.add(component)


@dataclass
class Settings:
    """
    General customization of a model.
        imperial_units (bool):
            True for imperial <3:
                in, lb, lb/(in/s2)
            False for SI:
                m, N, kg
        ndm, ndf: change them to break the code.
    """
    imperial_units: bool = field(default=True)  # false for SI
    ndm: int = field(default=3)  # that's all we support
    ndf: int = field(default=6)  # that's all we support

    def __repr__(self):
        res = ''
        res += '~~~ Model Settings ~~~\n'
        res += f'  Imperial units: {self.imperial_units}\n'
        res += f'  ndm           : {self.ndm}\n'
        res += f'  ndf           : {self.ndf}\n'
        return res


@dataclass(repr=False)
class Model:
    """
    Model object.
    Attributes:
        levels (CollectionActive)
        elastic_sections (Collection)
        fiber_sections (Collection)
        physical_materials (PhysicalMaterialCollection)
        component_connectivity (dict[tuple[int, ...], int])
        uid_generator (UIDGenerator)
        settings
    """
    name: str
    levels: collections.CollectionActive[int, Level] = field(
        init=False)
    elastic_sections: collections.Collection[int, ElasticSection] = field(
        init=False)
    fiber_sections: collections.Collection[int, FiberSection] = field(
        init=False)
    uniaxial_materials: collections.Collection[int, UniaxialMaterial] = field(
        init=False)
    physical_materials: collections.Collection[int, PhysicalMaterial] = field(
        init=False)
    uid_generator: UIDGenerator = field(
        default_factory=UIDGenerator)
    settings: Settings = field(default_factory=Settings)

    def __post_init__(self):
        self.levels = collections.CollectionActive(self)
        self.elastic_sections = collections.Collection(self)
        self.fiber_sections = collections.Collection(self)
        self.uniaxial_materials = collections.Collection(self)
        self.physical_materials = collections.Collection(self)

    def __repr__(self):
        res = ''
        res += '~~~ Model Object ~~~\n'
        res += f'ID: {id(self)}\n'
        res += f'levels: {self.levels.__srepr__()}\n'
        res += f'elastic_sections: {self.elastic_sections.__srepr__()}\n'
        res += f'fiber_sections: {self.fiber_sections.__srepr__()}\n'
        res += f'uniaxial_materials: {self.uniaxial_materials.__srepr__()}\n'
        res += f'physical_materials: {self.physical_materials.__srepr__()}\n'
        return res

    def component_connectivity(self) -> dict[tuple[int, ...],
                                             ComponentAssembly]:
        """
        Returns the connectivity of all component
        assemblies. Component assemblies are collections of
        lower-level components that are connected to primary
        nodes. Each component assembly can be represented by a typle
        of node uids of its connected nodes in ascending order. This
        method returns a dictionary having these tuples as keys, and
        the associated components as values.
        """
        res = {}
        components = self.list_of_components()
        for component in components:
            uids = [node.uid for node in component.external_nodes.values()]
            uids.sort()
            uids_tuple = (*uids,)
            assert uids_tuple not in res, 'Error! Duplicate component found.'
            res[uids_tuple] = component
        return res

    def add_level(self,
                  uid: int,
                  elevation: float):
        """
        Adds a level to the model.
        """
        lvl = Level(self, uid=uid, elevation=elevation)
        self.levels.add(lvl)

    def dict_of_primary_nodes(self):
        """
        Returns a dictionary of all the primary nodes in the model.
        The keys are the uids of the nodes.
        """
        dict_of_nodes: dict[int, Node] = {}
        for lvl in self.levels.values():
            dict_of_nodes.update(lvl.nodes)
        return dict_of_nodes

    def list_of_primary_nodes(self):
        """
        Returns a list of all the primary nodes in the model.
        """
        list_of_nodes = []
        for lvl in self.levels.values():
            for node in lvl.nodes.values():
                list_of_nodes.append(node)
        return list_of_nodes

    def dict_of_internal_nodes(self):
        """
        Returns a dictionary of all the internal nodes in the model.
        The keys are the uids of the nodes.
        """
        dict_of_nodes: dict[int, Node] = {}
        for lvl in self.levels.values():
            for component in lvl.components.values():
                dict_of_nodes.update(component.internal_nodes)
        return dict_of_nodes

    def list_of_internal_nodes(self):
        """
        Returns a list of all the internal nodes in the model.
        """
        list_of_nodes = []
        for lvl in self.levels.values():
            for component in lvl.components.values():
                for inode in component.internal_nodes.values():
                    list_of_nodes.append(inode)
        return list_of_nodes

    def dict_of_all_nodes(self):
        """
        Returns a dictionary of all the nodes in the model.
        The keys are the uids of the nodes.
        """
        dict_of_nodes: dict[int, Node] = {}
        dict_of_nodes.update(self.dict_of_primary_nodes())
        dict_of_nodes.update(self.dict_of_internal_nodes())
        return dict_of_nodes

    def list_of_all_nodes(self):
        """
        Returns a list of all the nodes in the model.
        """
        list_of_nodes = []
        list_of_nodes.extend(self.list_of_primary_nodes())
        list_of_nodes.extend(self.list_of_internal_nodes())
        return list_of_nodes

    def dict_of_components(self):
        """
        Returns a dictionary of all the component assemblies in the
        model.
        The keys are the uids of the component assemblies.
        """
        comps: dict[int, object] = {}
        for lvl in self.levels.values():
            for component in lvl.components.values():
                comps[component.uid] = component
        return comps

    def list_of_components(self):
        """
        Returns a list of all the component assembiles in the
        model.
        """
        return list(self.dict_of_components().values())

    def dict_of_elastic_beamcolumn_elements(self):
        """
        Returns a dictionary of all ElasticBeamColumn objects in the model.
        The keys are the uids of the objects.
        """
        elems: dict[int, ElasticBeamColumn] = {}
        for lvl in self.levels.values():
            for component in lvl.components.values():
                elems.update(component.elastic_beamcolumn_elements)
        return elems

    def list_of_elastic_beamcolumn_elements(self):
        """
        Returns a list of all ElasticBeamColumn objects in the model.
        """
        return list(self.dict_of_elastic_beamcolumn_elements().values())

    def dict_of_disp_beamcolumn_elements(self):
        """
        Returns a dictionary of all DispBeamColumn objects in the model.
        The keys are the uids of the objects.
        """
        elems: dict[int, DispBeamColumn] = {}
        for lvl in self.levels.values():
            for component in lvl.components.values():
                elems.update(component.disp_beamcolumn_elements)
        return elems

    def list_of_disp_beamcolumn_elements(self):
        """
        Returns a list of all DispBeamColumn objects in the model.
        """
        return list(self.dict_of_disp_beamcolumn_elements().values())

    def dict_of_beamcolumn_elements(self):
        """
        Returns a dictionary of all beamcolumn elements in the model.
        The keys are the uids of the objects.
        """
        elems = {}
        elems.update(self.dict_of_elastic_beamcolumn_elements())
        elems.update(self.dict_of_disp_beamcolumn_elements())
        return elems

    def list_of_beamcolumn_elements(self):
        """
        Returns a list of all beamcolumn elements in the model.
        """
        elems = []
        elems.extend(self.list_of_elastic_beamcolumn_elements())
        elems.extend(self.list_of_disp_beamcolumn_elements())
        return elems

    def dict_of_zerolength_elements(self):
        """
        Returns a dictionary of all zerolength elements in the model.
        The keys are the uids of the objects.
        """
        elems: dict[int, ZeroLength] = {}
        for lvl in self.levels.values():
            for component in lvl.components.values():
                elems.update(component.zerolength_elements)
        return elems

    def list_of_zerolength_elements(self):
        """
        Returns a list of all zerolength elements in the model.
        """
        return list(self.dict_of_zerolength_elements().values())

    def reference_length(self):
        """
        Returns the largest dimension of the
        bounding box of the building
        (used in graphics)
        """
        p_min = np.full(3, np.inf)
        p_max = np.full(3, -np.inf)
        for node in self.list_of_primary_nodes():
            point: nparr = np.array(node.coords)
            p_min = np.minimum(p_min, point)
            p_max = np.maximum(p_max, point)
        ref_len = np.max(p_max - p_min)
        return ref_len

    def initialize_empty_copy(self, name):
        """
        Initializes a shallow empty copy of the model.
        Used to create subset models.
        """
        res = Model(name)
        # copy the settings attributes
        res.settings.imperial_units = self.settings.imperial_units
        res.settings.ndf = self.settings.ndf
        res.settings.ndm = self.settings.ndm
        # make a copy of the levels
        for lvlkey, lvl in self.levels.items():
            res.add_level(lvlkey, lvl.elevation)
        # giv access to the materials and sections
        res.elastic_sections = self.elastic_sections
        res.fiber_sections = self.fiber_sections
        res.physical_materials = self.physical_materials
        res.uniaxial_materials = self.uniaxial_materials
        return res

    def transfer_by_polygon_selection(
            self,
            other: Model,
            coords: nparr):
        """
        Uses `transfer_component` to transfer all components of which
        the projection to the XY plane falls inside the specified
        polygon.
        """
        all_components = self.list_of_components()
        selected_components = []
        shape = shapely_Polygon(coords)
        for component in all_components:
            accept = True
            nodes = component.external_nodes.values()
            for node in nodes:
                if not shape.contains(Point(node.coords[0:2])):
                    accept = False
                    break
            if accept:
                selected_components.append(component)
        for component in selected_components:
            transfer_component(other, component)



"""
Model Generator for OpenSees ~ section
"""

#                          __
#   ____  ____ ___  ____ _/ /
#  / __ \/ __ `__ \/ __ `/ /
# / /_/ / / / / / / /_/ /_/
# \____/_/ /_/ /_/\__, (_)
#                /____/
#
# https://github.com/ioannis-vm/OpenSees_Model_Generator

from __future__ import annotations
from typing import TYPE_CHECKING
from typing import Optional
from typing import Any
from dataclasses import dataclass, field
import numpy as np
import numpy.typing as npt
from ..mesh import Mesh
from ..mesh import polygon_area
from .. import common
if TYPE_CHECKING:
    from .uniaxial_material import UniaxialMaterial
    from ..physical_material import PhysicalMaterial

nparr = npt.NDArray[np.float64]


@dataclass
class Section:
    """
    Section object.
    The axes are defined in the same way as they are
    defined in OpenSees. The colors assigned to
    the axes for plotting follow the
    AutoCAD convention.

            y(green)
            ^         x(red)
            :       .
            :     .
            :   .
           ===
            | -------> z (blue)
           ===

    """
    name: str
    uid: int


@dataclass
class ElasticSection(Section):
    """
    Elastic Section Object
    """
    e_mod: float
    area: float
    i_y: float
    i_x: float
    g_mod: float
    j_mod: float
    sec_w: float
    outside_shape: Optional[Mesh] = field(default=None, repr=False)
    snap_points: Optional[dict[str, nparr]] = field(default=None, repr=False)
    properties: Optional[dict[str, Any]] = field(default=None, repr=False)

    def weight_per_length(self):
        """
        Returns the weight per length of a section.
        For steel W sections, it adds 15% for misc. steel and connections.
        """
        if self.name[0] == 'W':
            res = self.sec_w * 1.15  # misc steel and connections
        else:
            res = self.sec_w
        return res

    def __repr__(self):
        res = ''
        res += 'ElasticSection object\n'
        res += f'name: {self.name}\n'
        res += f'uid: {self.uid}\n'
        res += 'Properties:'
        res += f'  E: {self.e_mod}\n'
        res += f'  A: {self.area}\n'
        res += f'  Iy: {self.i_y}\n'
        res += f'  Ix: {self.i_x}\n'
        res += f'  G: {self.g_mod}\n'
        res += f'  J: {self.j_mod}\n'
        res += f'  W: {self.sec_w}\n'
        if self.outside_shape:
            res += 'outside_shape: specified\n'
        else:
            res += 'outside_shape: None\n'
        if self.snap_points:
            res += 'snap_points: specified\n'
        else:
            res += 'snap_points: None\n'
        return res


@dataclass(repr=False)
class SectionComponent:
    """
    Part of a section object, having a single material.
    """
    outside_shape: Mesh
    holes: dict[str, Mesh]
    ops_material: UniaxialMaterial
    physical_material: PhysicalMaterial

    def __repr__(self):
        res = ''
        res += 'SectionComponent object\n'
        if self.outside_shape:
            res += 'outside_shape: specified\n'
        else:
            res += 'outside_shape: None\n'
        if self.holes:
            res += 'holes: exist\n'
        else:
            res += 'holes: no holes\n'
        res += f'ops_material: {self.ops_material.name}\n'
        res += f'physical_material: {self.physical_material.name}\n'
        return res


@dataclass(repr=False)
class FiberSection(Section):
    """
    Fiber section object.
    Can consist of multiple materials.
    The primary part of the component must have the key 'main'.
    """
    outside_shape: Mesh
    section_parts: dict[str, SectionComponent]
    j_mod: float
    snap_points: Optional[dict[str, nparr]] = field(default=None, repr=False)
    n_x: int = field(default=10)  # TODO: this shoule be editable
    n_y: int = field(default=10)

    def __repr__(self):
        res = ''
        res += 'FiberSection object\n'
        for part in self.section_parts:
            res += part.__repr__()
        if self.snap_points:
            res += 'snap_points: specified\n'
        else:
            res += 'snap_points: None\n'
        res += f'n_x: {self.n_x}, n_y: {self.n_y}\n'
        return res

    def ops_args(self):
        """
        Returns the arguments required to define the object in
        OpenSees
        """
        return ['Fiber', self.uid, '-GJ',
                self.j_mod*self.section_parts['main'].physical_material.g_mod]

    def weight_per_length(self):
        """
        Returns the weight per length of a section.
        For steel W sections, it adds 15% for misc. steel and connections.
        """
        if self.name[0] == 'W':
            mult = 1.15  # misc steel and connections
        else:
            mult = 1.00
        res = 0.00
        for part in self.section_parts.values():
            coords: nparr = np.array(
                [h.vertex.coords for h in part.outside_shape.halfedges])
            area = polygon_area(coords)
            for hole in part.holes:
                hole_coords: nparr = np.array(
                    [h.vertex.coords
                     for h in part.holes[hole].halfedges])
                area -= polygon_area(hole_coords)
            density = part.physical_material.density
            # TODO: units
            res += area * density * common.G_CONST_IMPERIAL
        return res * mult



"""
Model Generator for OpenSees ~ element
"""

#                          __
#   ____  ____ ___  ____ _/ /
#  / __ \/ __ `__ \/ __ `/ /
# / /_/ / / / / / / /_/ /_/
# \____/_/ /_/ /_/\__, (_)
#                /____/
#
# https://github.com/ioannis-vm/OpenSees_Model_Generator

# pylint: disable=invalid-name


from dataclasses import dataclass


@dataclass
class UniaxialMaterial:
    """
    OpenSees uniaxialMaterial
    https://openseespydoc.readthedocs.io/en/latest/src/uniaxialMaterial.html
    """
    uid: int
    name: str


@dataclass
class Elastic(UniaxialMaterial):
    """
    OpenSees Elastic
    https://openseespydoc.readthedocs.io/en/latest/src/ElasticUni.html
    """
    e_mod: float

    def ops_args(self):
        """
        Returns the arguments required to define the object in
        OpenSees
        """
        return [
            'Elastic',
            self.uid,
            self.e_mod
        ]


@dataclass
class Steel02(UniaxialMaterial):
    """
    OpenSees Steel02
    https://openseespydoc.readthedocs.io/en/latest/src/steel02.html
    """
    Fy: float
    E0: float
    b: float
    params: tuple[float, float, float]
    a1: float
    a2: float
    a3: float
    a4: float
    sigInit: float
    G: float

    def ops_args(self):
        """
        Returns the arguments required to define the object in
        OpenSees
        """
        return [
            'Steel02',
            self.uid,
            self.Fy,
            self.E0,
            self.b,
            *self.params,
            self.a1,
            self.a2,
            self.a3,
            self.a4,
            self.sigInit
        ]


@dataclass
class Bilin(UniaxialMaterial):
    """
    OpenSees Bilin Material
    https://openseespydoc.readthedocs.io/en/latest/src/Bilin.html
    """
    K0: float
    as_Plus: float
    as_Neg: float
    My_Plus: float
    My_Neg: float
    Lamda_S: float
    Lamda_C: float
    Lamda_A: float
    Lamda_K: float
    c_S: float
    c_C: float
    c_A: float
    c_K: float
    theta_p_Plus: float
    theta_p_Neg: float
    theta_pc_Plus: float
    theta_pc_Neg: float
    Res_Pos: float
    Res_Neg: float
    theta_u_Plus: float
    theta_u_Neg: float
    D_Plus: float
    D_Neg: float
    nFactor: float

    def ops_args(self):
        """
        Returns the arguments required to define the object in
        OpenSees
        """
        return [
            'Bilin',
            self.uid,
            self.K0,
            self.as_Plus,
            self.as_Neg,
            self.My_Plus,
            self.My_Neg,
            self.Lamda_S,
            self.Lamda_C,
            self.Lamda_A,
            self.Lamda_K,
            self.c_S,
            self.c_C,
            self.c_A,
            self.c_K,
            self.theta_p_Plus,
            self.theta_p_Neg,
            self.theta_pc_Plus,
            self.theta_pc_Neg,
            self.Res_Pos,
            self.Res_Neg,
            self.theta_u_Plus,
            self.theta_u_Neg,
            self.D_Plus,
            self.D_Neg,
            self.nFactor
        ]


@dataclass
class Pinching4(UniaxialMaterial):
    """
    OpenSees Pinching4 Material
    https://openseespydoc.readthedocs.io/en/latest/src/Pinching4.html?highlight=pinching4
    """
    ePf1: float
    ePf2: float
    ePf3: float
    ePf4: float
    ePd1: float
    ePd2: float
    ePd3: float
    ePd4: float
    eNf1: float
    eNf2: float
    eNf3: float
    eNf4: float
    eNd1: float
    eNd2: float
    eNd3: float
    eNd4: float
    rDispP: float
    fForceP: float
    uForceP: float
    rDispN: float
    fFoceN: float
    uForceN: float
    gK1: float
    gK2: float
    gK3: float
    gK4: float
    gKLim: float
    gD1: float
    gD2: float
    gD3: float
    gD4: float
    gDLim: float
    gF1: float
    gF2: float
    gF3: float
    gF4: float
    gFLim: float
    gE: float
    dmgType: str

    def ops_args(self):
        """
        Returns the arguments required to define the object in
        OpenSees
        """
        return [
            'Pinching4',
            self.uid,
            self.ePf1, self.ePf2, self.ePf3, self.ePf4,
            self.ePd1, self.ePd2, self.ePd3, self.ePd4,
            self.eNf1, self.eNf2, self.eNf3, self.eNf4,
            self.eNd1, self.eNd2, self.eNd3, self.eNd4,
            self.rDispP, self.fForceP, self.uForceP,
            self.rDispN, self.fFoceN, self.uForceN,
            self.gK1, self.gK2, self.gK3, self.gK4, self.gKLim,
            self.gD1, self.gD2, self.gD3, self.gD4, self.gDLim,
            self.gF1, self.gF2, self.gF3, self.gF4, self.gFLim,
            self.gE, self.dmgType
        ]


@dataclass
class Hysteretic(UniaxialMaterial):
    """
    OpenSees Bilin Material
    https://openseespydoc.readthedocs.io/en/latest/src/Bilin.html
    """
    p1: tuple[float, float]
    p2: tuple[float, float]
    p3: tuple[float, float]
    n1: tuple[float, float]
    n2: tuple[float, float]
    n3: tuple[float, float]
    pinchX: float
    pinchY: float
    damage1: float
    damage2: float
    beta: float

    def ops_args(self):
        """
        Returns the arguments required to define the object in
        OpenSees
        """
        return [
            'Hysteretic',
            self.uid,
            *self.p1,
            *self.p2,
            *self.p3,
            *self.n1,
            *self.n2,
            *self.n3,
            self.pinchX,
            self.pinchY,
            self.damage1,
            self.damage2,
            self.beta
        ]



"""
Model Generator for OpenSees ~ steel design checks
"""

#                          __
#   ____  ____ ___  ____ _/ /
#  / __ \/ __ `__ \/ __ `/ /
# / /_/ / / / / / / /_/ /_/
# \____/_/ /_/ /_/\__, (_)
#                /____/
#
# https://github.com/ioannis-vm/OpenSees_Model_Generator

from __future__ import annotations
from typing import Optional


def smrf_scwb(
        col_sec_properties: dict[str, float],
        col_sec_properties_above: dict[str, float],
        beam_i_sec_properties: dict[str, float],
        col_axial_load: float,
        beam_udl_i: float,
        rbs_proportion_i: float,
        level_height: float,
        bay_length: float,
        beam_j_sec_properties: Optional[dict[str, float]],
        beam_udl_j: Optional[float],
        rbs_proportion_j: Optional[float],
        s_h: float,
        f_y,
        ry_coeff=1.15,
        omega_coeff=1.10,
):
    """
    SMRF strong column weak beam check
    """
    # in3
    zc_below = col_sec_properties['Zx']
    # in2
    ac_below = col_sec_properties['A']
    # in3
    zc_above = col_sec_properties_above['Zx']
    # in2
    ac_above = col_sec_properties_above['A']
    # lb-in
    mc_below = zc_below * (f_y - col_axial_load/ac_below)
    # lb-in
    mc_above = zc_above * (f_y - col_axial_load/ac_above)
    # lb
    if beam_j_sec_properties:
        max_beam_d = max(beam_i_sec_properties['d'],
                         beam_j_sec_properties['d'])
    else:
        max_beam_d = beam_i_sec_properties['d']
    vc_star = (mc_below + mc_above) / ((level_height - max_beam_d))
    # lb-in
    sigma_mc_star = ((mc_below + mc_above)
                     + vc_star * max_beam_d/2.00)
    # in
    c_rbs_i = beam_i_sec_properties['bf'] * (1. - rbs_proportion_i) / 2.
    # in3
    z_rbs_i = (beam_i_sec_properties['Zx']
               - 2. * c_rbs_i * beam_i_sec_properties['tf']
               * (beam_i_sec_properties['d']
                  - beam_i_sec_properties['tf']))
    # lb-in
    m_pr_i = ry_coeff * omega_coeff * f_y * z_rbs_i
    # lb
    v_e_i = (2 * m_pr_i) / (bay_length - 2. * s_h)
    # lb
    v_g_i = beam_udl_i * (bay_length - 2. * s_h) / 2.
    d_c = col_sec_properties['d']

    if beam_j_sec_properties:
        assert beam_udl_j is not None
        assert rbs_proportion_j is not None
        c_rbs_j = beam_j_sec_properties['bf'] * (1. - rbs_proportion_j) / 2.
        z_rbs_j = (beam_j_sec_properties['Zx']
                   - 2. * c_rbs_j * beam_j_sec_properties['tf']
                   * (beam_j_sec_properties['d']
                      - beam_j_sec_properties['tf']))
        m_pr_j = ry_coeff * omega_coeff * f_y * z_rbs_j
        v_e_j = (2 * m_pr_j) / (bay_length - 2. * s_h)
        v_g_j = beam_udl_j * (bay_length - 2. * s_h) / 2.

        sigm_mb_star = (m_pr_i + v_e_i * (s_h + d_c/2.) + v_g_i*(s_h + d_c/2.)
                        + m_pr_j + v_e_j * (s_h + d_c/2.)
                        - v_g_j*(s_h + d_c/2.))
        capacity = sigma_mc_star / sigm_mb_star

    else:
        sigm_mb_star = 1.00 * (m_pr_i + v_e_i * (s_h + d_c/2.)
                               + v_g_i*(s_h + d_c/2.))
        capacity = sigma_mc_star / sigm_mb_star

    return capacity


def smrf_pz_doubler_plate_requirement(
        col_sec_properties: dict[str, float],
        beam_sec_properties: dict[str, float],
        rbs_proportion: float,
        bay_length: float,
        place: str,
        s_h: float,
        f_y,
        ry_coeff=1.15,
        omega_coeff=1.10,
):
    """
    Calculates the required doubler plate thickness
    """
    # in
    c_rbs = beam_sec_properties['bf'] * (1. - rbs_proportion) / 2.
    # in3
    z_rbs = (beam_sec_properties['Zx']
             - 2. * c_rbs * beam_sec_properties['tf']
             * (beam_sec_properties['d'] - beam_sec_properties['tf']))
    # lb-in
    m_pr = ry_coeff * omega_coeff * f_y * z_rbs
    # lb
    v_e = 2. * m_pr / (bay_length - 2. * s_h)
    # lb-in
    m_f = m_pr + v_e * s_h
    # lb
    r_n = (0.60 * f_y * col_sec_properties['d'] * col_sec_properties['tw']
           * (1.00 + (3. * col_sec_properties['bf']
                      * (col_sec_properties['tf'])**2)
              / (col_sec_properties['d']
                 * beam_sec_properties['d'] * col_sec_properties['tw'])))
    if place == 'interior':
        r_u = 2 * m_f / (beam_sec_properties['d'] - beam_sec_properties['tf'])
        tdoub = (r_u-r_n) / (0.60 * f_y * col_sec_properties['d'])
        tdoub = max(tdoub, 0.00)
    else:
        r_u = m_f / (beam_sec_properties['d'] - beam_sec_properties['tf'])
        tdoub = (r_u-r_n) / (0.60 * f_y * col_sec_properties['d'])
        tdoub = max(tdoub, 0.00)
    return tdoub


# def steel_W_sec_strength_check(sec_properties, loads):
#     pass



"""
Model Generator for OpenSees ~ component assembly
"""

#                          __
#   ____  ____ ___  ____ _/ /
#  / __ \/ __ `__ \/ __ `/ /
# / /_/ / / / / / / /_/ /_/
# \____/_/ /_/ /_/\__, (_)
#                /____/
#
# https://github.com/ioannis-vm/OpenSees_Model_Generator

from __future__ import annotations
from dataclasses import dataclass, field
import numpy as np
import numpy.typing as npt
from . import collections
from .ops import element


nparr = npt.NDArray[np.float64]


@dataclass
class ComponentAssembly:
    """
    A component assembly represents some part of a structure and holds
    various lower-level elements such as nodes and beamcolumn
    elements.
    Attributes:
      uid (int): Unique identifyer of the component assembly
      parent_collection (ComponentCollection): The collection of
        elements to which the component assembly belongs.
      component_purpose (str): The functionality of the component assembly
      external_nodes (NodeCollection): the external nodes to which the
        component assembly is connected.
        these nodes should exist as part of a level.
      internal_nodes (NodeCollection): internal nodes that are
        required for the connectivity of the elements of the component
        assembly.
        these nodes only exist as part of the component assembly.
      elastic_beamcolumn_elements (CollectionWithConnectivity): ...
      disp_beamcolumn_elements (CollectionWithConnectivity): ...
      zerolength_elements (CollectionWithConnectivity): ...
    """
    uid: int
    parent_collection: collections.Collection[int, ComponentAssembly]
    component_purpose: str
    external_nodes: collections.NodeCollection = field(
        init=False)
    internal_nodes: collections.NodeCollection = field(
        init=False)
    elastic_beamcolumn_elements: (collections.CollectionWithConnectivity[
        int, element.ElasticBeamColumn]) = field(init=False)
    disp_beamcolumn_elements: (collections.CollectionWithConnectivity[
        int, element.DispBeamColumn]) = field(init=False)
    zerolength_elements: (collections.CollectionWithConnectivity[
        int, element.ZeroLength]) = field(init=False)

    def __post_init__(self):
        self.external_nodes = collections.NodeCollection(self)
        self.internal_nodes = collections.NodeCollection(self)
        self.elastic_beamcolumn_elements = \
            collections.CollectionWithConnectivity(self)
        self.disp_beamcolumn_elements = \
            collections.CollectionWithConnectivity(self)
        self.zerolength_elements = collections.CollectionWithConnectivity(self)

    def __srepr__(self):
        """
        Short version of repr
        """
        return f'Component assembly, uid: {self.uid}'

    def __repr__(self):
        res = ''
        res += 'Component assembly object\n'
        res += f'uid: {self.uid}\n'
        res += f'component_purpose: {self.component_purpose}\n'
        res += 'External Nodes\n'
        for node in self.external_nodes.values():
            res += f'  {node.uid}, {node.coords}'
        res += 'Internal Nodes\n'
        for node in self.internal_nodes.values():
            res += f'  {node.uid}, {node.coords}'
        return res

    def dict_of_elastic_beamcolumn_elements(self):
        """
        Returns a dictionary of all ElasticBeamColumn objects in the model.
        The keys are the uids of the objects.
        """
        res = {}
        for elm in self.elastic_beamcolumn_elements.values():
            res[elm.uid] = elm
        return res

    def list_of_elastic_beamcolumn_elements(self):
        """
        Returns a list of all ElasticBeamColumn objects in the model.
        """
        return list(self.dict_of_elastic_beamcolumn_elements().values())

    def dict_of_disp_beamcolumn_elements(self):
        """
        Returns a dictionary of all DispBeamColumn objects in the model.
        The keys are the uids of the objects.
        """
        res = {}
        for elm in self.disp_beamcolumn_elements.values():
            res[elm.uid] = elm
        return res

    def list_of_disp_beamcolumn_elements(self):
        """
        Returns a list of all DispBeamColumn objects in the model.
        """
        return list(self.dict_of_disp_beamcolumn_elements().values())

    def dict_of_beamcolumn_elements(self):
        """
        Returns a dictionary of all beamcolumn elements in the model.
        The keys are the uids of the objects.
        """
        res = {}
        res.update(self.dict_of_elastic_beamcolumn_elements())
        res.update(self.dict_of_disp_beamcolumn_elements())
        return res

    def list_of_all_elements(self):
        """
        Returns a list of all beamcolumn elements in the model.
        """
        return list(self.dict_of_beamcolumn_elements().values())

    def element_connectivity(self):
        """
        Returns the connectivity of all elements. Elements are
        connected to external nodes. Each component assembly can be
        represented by a typle of node uids of its connected nodes in
        ascending order. This method returns a dictionary having these
        tuples as keys, and the associated components as values.
        """
        res = {}
        elms = self.list_of_all_elements()
        for elm in elms:
            uids = [nd.uid for nd in elm.nodes]
            uids.sort()
            uids_tuple = (*uids,)
            res[uids_tuple] = elm
        return res



"""
Model Generator for OpenSees ~ solver
"""

#                          __
#   ____  ____ ___  ____ _/ /
#  / __ \/ __ `__ \/ __ `/ /
# / /_/ / / / / / / /_/ /_/
# \____/_/ /_/ /_/\__, (_)
#                /____/
#
# https://github.com/ioannis-vm/OpenSees_Model_Generator

# pylint: disable=no-member


from __future__ import annotations
from typing import Optional
from typing import Any
from typing import TYPE_CHECKING
from dataclasses import dataclass, field
import os
import pickle
import logging
import time
import numpy as np
import numpy.typing as npt
from scipy import integrate  # type: ignore
from scipy.interpolate import interp1d  # type: ignore
import plotly.express as px
import pandas as pd
import openseespy.opensees as ops  # type: ignore
import matplotlib.pyplot as plt  # type: ignore
from .load_case import LoadCase
from .mesh import subdivide_polygon
# from .import components
from .model import Model
from .ops.element import ElasticBeamColumn
from .import common
# from .graphics import postprocessing_3d
from .graphics import general_2d
from .import transformations
from .collections import Collection
from .gen.querry import LoadCaseQuerry
if TYPE_CHECKING:
    from .ops.uniaxial_material import UniaxialMaterial

nparr = npt.NDArray[np.float64]


plt.rc('font', family='serif')
plt.rc('xtick', labelsize='medium')
plt.rc('ytick', labelsize='medium')

LN_ANALYSIS_SYSTEM = 'SparseSYM'
NL_ANALYSIS_SYSTEM = 'SparseSYM'
MD_ANALYSIS_SYSTEM = 'UmfPack'
CONSTRAINTS = ('Transformation',)
NUMBERER = 'Plain'


def test_uniaxial_material(
        mat: UniaxialMaterial,
        input_deformations: list[float],
        incr: float,
        plot: bool
):
    """
    Generates force-deformation pairs using the given uniaxial material.
    Used for testing and validation.
    """

    forces = [0.00]
    deformations = [0.00]

    # define model
    ops.wipe()
    ops.model('basic', '-ndm', 1, '-ndf', 1)
    ops.node(0, 0.00)
    ops.node(1, 0.00)
    ops.fix(0, 1)
    ops.uniaxial_material(  # type: ignore
        *mat.ops_args())  # type: ignore
    ops.element('zeroLength', 0, 0, 1, '-mat', mat.uid, '-dir', 1)
    # define load pattern
    ops.timeSeries('Linear', 1)
    ops.pattern('Plain', 1, 1)
    ops.load(1, 1.00)
    # run analysis
    ops.numberer('Plain')
    ops.constraints('Plain')

    curr_defo = ops.nodeDisp(1, 1)
    n_steps_success = 0
    total_fail = False
    num_subdiv = 0
    num_times = 0
    scale = [1.0, 0.1, 0.01, 0.001, 0.0001, 0.00001]
    steps = [25, 50, 75, 100, 250, 500]
    norm = [1.0e-6] * 5 + [1.0e-2]
    algor = ['Newton']*6
    for target_defo in input_deformations:
        # determine push direction
        if curr_defo < target_defo:
            incr = abs(incr)
            sign = +1.00
        else:
            incr = -abs(incr)
            sign = -1.00
        print('entering loop', target_defo)
        while curr_defo * sign < target_defo * sign:
            # determine increment
            if abs(curr_defo - target_defo) < \
               abs(incr) * scale[num_subdiv]:
                incr_anl = sign * abs(curr_defo - target_defo)
            else:
                incr_anl = incr * scale[num_subdiv]
            ops.test('NormDispIncr', norm[num_subdiv],
                     steps[num_subdiv], 0)
            ops.algorithm(algor[num_subdiv])
            # ops.integrator("ArcLength", 1.00e1, 1.00e-7)
            ops.integrator("DisplacementControl", 1, 1, incr_anl)
            ops.system('FullGeneral')
            ops.analysis("Static")
            flag = ops.analyze(1)
            if flag != 0:
                # analysis failed
                if num_subdiv == len(scale) - 1:
                    # can't refine further
                    print()
                    print('===========================')
                    print('Analysis failed to converge')
                    print('===========================')
                    print()
                    total_fail = True
                    break
                # can still reduce step size
                num_subdiv += 1
                # how many times to run with reduced step size
                num_times = 10
            else:
                # analysis was successful
                n_steps_success += 1
                if num_times != 0:
                    num_times -= 1
                if num_subdiv != 0:
                    if num_times == 0:
                        num_subdiv -= 1
                        num_times = 10
                ops.reactions()
                curr_defo = ops.nodeDisp(1, 1)
                reaction = -ops.nodeReaction(0)[0]
                forces.append(reaction)
                deformations.append(curr_defo)
            if total_fail:
                break
    print('Number of saved analysis steps:', n_steps_success)
    results: nparr = np.column_stack((deformations, forces))
    dframe = pd.DataFrame(results)
    dframe.columns = ['deformation', 'froce']
    if plot:
        fig = px.line(dframe, x='deformation', y='force', markers=True)
        fig.show()
    return dframe


@dataclass(repr=False)
class Results:
    """
    Results object. Stores analysis results.
    """
    node_displacements: Collection[
        int, dict[int, list[float]]] = field(init=False)
    node_velocities: Collection[
        int, dict[int, list[float]]] = field(init=False)
    node_accelerations: Collection[
        int, dict[int, list[float]]] = field(init=False)
    node_reactions: Collection[
        int, dict[int, list[float]]] = field(init=False)
    element_forces: Collection[
        int, dict[int, nparr]] = field(init=False)
    # fiber_stress_strain: Collection = field(init=False)
    release_force_defo: Collection[
        int, dict[int, list[float]]] = field(init=False)
    periods: Optional[nparr] = field(default=None)
    n_steps_success: Optional[int] = field(default=None)
    metadata: Optional[dict[str, object]] = field(default=None)

    def __post_init__(self):
        self.node_displacements = Collection(self)
        self.node_velocities = Collection(self)
        self.node_accelerations = Collection(self)
        self.node_reactions = Collection(self)
        self.element_forces = Collection(self)
        # self.fiber_stress_strain = Collection(self)
        self.release_force_defo = Collection(self)


@dataclass(repr=False)
class AnalysisStorageSettings:
    """
    Analysis storage settings object.
    Controls what will be stored and how.
    """
    log_file: Optional[str] = field(default=None)
    store_forces: bool = field(default=True)
    store_reactions: bool = field(default=True)
    store_fiber: bool = field(default=True)
    store_release_force_defo: bool = field(default=True)
    specific_nodes: list[int] = field(default_factory=list)
    pickle_results: bool = field(default=False)


@dataclass(repr=False)
class Analysis:
    """
    """
    mdl: Model
    load_cases: dict[str, LoadCase]
    output_directory: Optional[str] = field(default=None)
    silent: bool = field(default=False)
    settings: AnalysisStorageSettings = field(
        default_factory=AnalysisStorageSettings)
    results: dict[str, Results] = field(default_factory=dict)
    logger: Optional[object] = field(default=None)

    def log(self, msg: str):
        """
        Adds a message to the log file
        """
        if self.settings.log_file:
            self.logger.info(msg)  # type: ignore

    def print(self, thing: Any):
        """
        Prints a message to stdout
        """
        if not self.silent:
            print(thing)

    def _init_results(self):

        # initialize logger
        if self.settings.log_file:
            logging.basicConfig(
                filename=self.settings.log_file,
                format='%(asctime)s %(message)s',
                datefmt='%m/%d/%Y %I:%M:%S %p')
            self.logger = logging.getLogger('OpenSees_Model_Generator')
            self.logger.setLevel(logging.DEBUG)

        # initialize output directory
        if self.output_directory and not os.path.exists(self.output_directory):
            os.makedirs(
                self.output_directory,
                exist_ok=True)

        if self.settings.pickle_results and not self.output_directory:
            raise ValueError('Speficy an output directory for the results.')

        # initialize result collections
        assert isinstance(self.load_cases, dict)
        for case_name in self.load_cases:
            node_uids = []
            if self.settings.specific_nodes:
                node_uids.extend(self.settings.specific_nodes)
            else:
                node_uids.extend(nd.uid for nd in self.mdl.list_of_all_nodes())
                node_uids.extend(
                    [n.uid for n in self.load_cases[case_name]
                     .parent_nodes.values()])
            self.results[case_name] = Results()
            for uid in node_uids:
                self.results[case_name].node_displacements[uid] = {}
                self.results[case_name].node_velocities[uid] = {}
                self.results[case_name].node_accelerations[uid] = {}
                self.results[case_name].node_reactions[uid] = {}

            if self.settings.store_forces:
                for uid in self.mdl.dict_of_elastic_beamcolumn_elements():
                    self.results[case_name].element_forces[uid] = {}
            if self.settings.store_fiber:
                for uid in self.mdl.dict_of_disp_beamcolumn_elements():
                    self.results[case_name].element_forces[uid] = {}
            if self.settings.store_release_force_defo:
                for uid in self.mdl.dict_of_zerolength_elements():
                    self.results[case_name].release_force_defo[uid] = {}

        self.log('Analysis started')

    def _write_results_to_disk(self):
        """
        Pickles the results
        """
        with open(f'{self.output_directory}/main_results.pcl',
                  'wb') as file:
            pickle.dump(self.results, file)

    def read_results_from_disk(self):
        """
        Reads back results from a pickle file
        """
        with open(f'{self.output_directory}/main_results.pcl',
                  'rb') as file:
            self.results = pickle.load(file)

    def _to_opensees_domain(self, case_name):
        """
        Defines the model in OpenSeesPy
        """

        # initialize
        ops.wipe()
        ops.model('basic', '-ndm', 3, '-ndf', 6)

        # ~~~~~~~~~~~~~~~ #
        # Node definition #
        # ~~~~~~~~~~~~~~~ #

        # keep track of defined nodes
        defined_nodes = {}

        primary_nodes = self.mdl.dict_of_primary_nodes()
        internal_nodes = self.mdl.dict_of_internal_nodes()
        parent_nodes = {}
        parent_node_to_lvl = {}
        for lvl_uid, node in self.load_cases[case_name].parent_nodes.items():
            parent_nodes[node.uid] = node
            parent_node_to_lvl[node.uid] = lvl_uid

        all_nodes = {}
        all_nodes.update(primary_nodes)
        all_nodes.update(internal_nodes)
        all_nodes.update(parent_nodes)
        for uid, node in all_nodes.items():
            if uid in defined_nodes:
                raise KeyError(f'Node already defined: {uid}')
            defined_nodes[uid] = node
            ops.node(node.uid, *node.coords)

        # restraints
        for uid, node in primary_nodes.items():
            ops.fix(node.uid, *node.restraint)
        for uid, node in internal_nodes.items():
            # (this is super unusual, but who knows..)
            ops.fix(node.uid, *node.restraint)
        for uid in parent_nodes:
            n_r = [False, False, True, True, True, False]
            ops.fix(uid, *n_r)

        # lumped nodal mass
        for uid, node in all_nodes.items():
            if np.max(np.array(
                    (*self.load_cases[case_name]
                     .node_mass[node.uid]
                     .val,))) < common.EPSILON:
                continue
            ops.mass(node.uid, *self.load_cases[case_name].node_mass
                     [node.uid].val)

        # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ #
        # Elastic BeamColumn element definition #
        # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ #

        # keep track of defined elements
        defined_elements = {}

        elms = self.mdl.dict_of_elastic_beamcolumn_elements().values()

        # define line elements
        for elm in elms:
            if elm.visibility.skip_opensees_definition:
                continue
            ops.geomTransf(*elm.geomtransf.ops_args())
            ops.element(*elm.ops_args())

        # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ #
        # Fiber BeamColumn element definition #
        # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ #

        # keep track of defined elements
        defined_sections = {}
        defined_materials = {}

        elms = self.mdl.dict_of_disp_beamcolumn_elements().values()

        for elm in elms:
            sec = elm.section
            parts = sec.section_parts.values()
            n_x = elm.section.n_x
            n_y = elm.section.n_y
            if sec.uid not in defined_sections:
                ops.section(*sec.ops_args())
                defined_sections[sec.uid] = sec
                for part in parts:
                    mat = part.ops_material
                    if mat.uid not in defined_materials:
                        ops.uniaxialMaterial(
                            *mat.ops_args())
                        defined_materials[mat.uid] = mat
                    pieces = subdivide_polygon(
                        part.outside_shape, part.holes, n_x, n_y)
                    for piece in pieces:
                        area = piece.area
                        z_loc = piece.centroid.x
                        y_loc = piece.centroid.y
                        ops.fiber(y_loc,
                                  z_loc,
                                  area,
                                  part.ops_material.uid)
                    plt.show()
            ops.beamIntegration(*elm.integration.ops_args())
            ops.geomTransf(*elm.geomtransf.ops_args())
            ops.element(*elm.ops_args())

        # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ #
        # ZeroLength element definition #
        # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ #

        elms = self.mdl.list_of_zerolength_elements()

        # define zerolength elements
        for elm in elms:
            for mat in elm.mats:
                if mat.uid not in defined_materials:
                    ops.uniaxialMaterial(
                        *mat.ops_args())
                    defined_materials[mat.uid] = mat
            ops.element(*elm.ops_args())
            defined_elements[elm.uid] = elm

        # node constraints
        for uid in parent_nodes:
            lvl = self.mdl.levels[parent_node_to_lvl[uid]]
            nodes = lvl.nodes.values()
            good_nodes = [n for n in nodes if n.coords[2] == lvl.elevation]
            ops.rigidDiaphragm(
                3, uid, *[nd.uid for nd in good_nodes])

    def _define_loads(self, case_name):
        ops.timeSeries('Linear', 1)
        ops.pattern('Plain', 1, 1)
        for elm in self.mdl.list_of_beamcolumn_elements():
            if elm.visibility.skip_opensees_definition:
                continue
            udl_total = (self.load_cases[case_name]
                         .line_element_udl[elm.uid].val)
            ops.eleLoad('-ele', elm.uid,
                        '-type', '-beamUniform',
                        udl_total[1],
                        udl_total[2],
                        udl_total[0])

        for node in self.mdl.list_of_all_nodes():
            ops.load(node.uid, *self.load_cases[case_name]
                     .node_loads[node.uid].val)

    ####################################################
    # Methods that read back information from OpenSees #
    ####################################################

    def _read_node_displacements(self, case_name, step, nodes):
        for node in nodes:
            if self.settings.specific_nodes:
                if node.uid not in self.settings.specific_nodes:
                    continue
            val = ops.nodeDisp(node.uid)
            self.results[case_name].node_displacements[node.uid][step] = val

    def _read_node_velocities(self, case_name, step, nodes):
        for node in nodes:
            if self.settings.specific_nodes:
                if node.uid not in self.settings.specific_nodes:
                    continue
            val = ops.nodeVel(node.uid)
            self.results[case_name].node_velocities[node.uid][step] = val

    def _read_node_accelerations(self, case_name, step, nodes):
        for node in nodes:
            if self.settings.specific_nodes:
                if node.uid not in self.settings.specific_nodes:
                    continue
            val = ops.nodeAccel(node.uid)
            self.results[case_name].node_accelerations[node.uid][step] = val

    def _read_node_reactions(self, case_name, step, nodes):
        ops.reactions()
        for node in nodes:
            if True in node.restraint:
                val = ops.nodeReaction(node.uid)
                self.results[case_name].node_reactions[node.uid][step] = val

    def _read_frame_element_forces(self, case_name, step, elems):
        for elm in elems:
            uid = elm.uid
            global_values: nparr = np.array(ops.eleForce(uid))
            forces_global = global_values[0:3]
            moments_global_ends = global_values[3:6]
            moments_global_clear = transformations.offset_transformation(
                elm.geomtransf.offset_i, moments_global_ends, forces_global)
            x_vec = elm.geomtransf.x_axis
            y_vec = elm.geomtransf.y_axis
            z_vec = elm.geomtransf.z_axis
            transf_global2local: nparr = np.vstack((x_vec, y_vec, z_vec))
            n_i, qy_i, qz_i = transf_global2local @ forces_global
            t_i, my_i, mz_i = transf_global2local @ moments_global_clear
            forces: nparr = np.array((n_i, qy_i, qz_i, t_i, my_i, mz_i))
            self.results[case_name].element_forces[uid][step] = forces

#     def _read_frame_fiber_stress_strain(self):
#         for elm in self.mdl.list_of_line_elements():
#             if elm.model_as['type'] != 'fiber':
#                 continue
#             uid = elm.uid
#             mat_id = elm.section.material.uid
#             result = []
#             n_p = elm.n_p
#             pts = elm.section.snap_points
#             for pt in pts.keys():
#                 pt = list(pts.keys())[0]
#                 z_loc = pts[pt][0]
#                 y_loc = pts[pt][1]
#                 stress_strain = []
#                 for i in range(n_p):
#                     stress_strain.append(ops.eleResponse(
#                         int(uid), "section", str(i+1), "-fiber", str(y_loc),
#                         str(z_loc), str(mat_id), "stressStrain"))
#                 result.append(stress_strain)
#             self._store_result(self.fiber_stress_strain, uid, result)

    def _read_release_moment_rot(self, case_name, step, zerolength_elms):
        for release in zerolength_elms:
            # force_global = ops.eleResponse(
            #     release.uid, 'force')[:3]
            moment_global = ops.eleResponse(
                release.uid, 'force')[3:6]
            # disp_local = ops.eleResponse(
            #     release.uid, 'deformation')[:3]
            rot_local = ops.eleResponse(
                release.uid, 'deformation')[3:6]
            # note: j quantities are the opposite of those of
            # i by equilibrium
            # no need to store them too
            # rotation_global = ops.eleResponse(
            #     release.uid, 'deformation')[3:6]
            # convert to the local system
            vec_x = release.vecx
            vec_y = release.vecyp
            vec_z: nparr = np.cross(vec_x, vec_y)
            tmat_g2l = transformations.transformation_matrix(
                vec_x, vec_y, vec_z)
            moment_local = tmat_g2l @ -(np.array(moment_global))
            self.results[case_name].release_force_defo[release.uid][step] = \
                [*rot_local, *moment_local]

    def _read_opensees_results(
            self,
            case_name,
            step,
            nodes,
            elastic_beamcolumn_elements,
            disp_beamcolumn_elements,
            zerolength_elements):
        self._read_node_displacements(case_name, step, nodes)
        self._read_node_velocities(case_name, step, nodes)
        self._read_node_accelerations(case_name, step, nodes)
        if self.settings.store_reactions:
            self._read_node_reactions(case_name, step, nodes)
        if self.settings.store_forces:
            self._read_frame_element_forces(
                case_name,
                step, elastic_beamcolumn_elements)
            self._read_frame_element_forces(
                case_name,
                step, disp_beamcolumn_elements)
        # if self.settings.store_fiber:
        #     self._read_frame_fiber_stress_strain()
        if self.settings.store_release_force_defo:
            self._read_release_moment_rot(
                case_name,
                step, zerolength_elements)

    ##################################
    # Numeric Result Post-processing #
    ##################################

    def global_reactions(self, case_name, step):
        """
        Calculates and returns the global reaction forces.
        """
        reactions = np.full(6, 0.00)
        for lvl in self.mdl.levels.values():
            for node in lvl.nodes.values():
                if True in node.restraint:
                    uid = node.uid
                    x_coord = node.coords[0]
                    y_coord = node.coords[1]
                    z_coord = node.coords[2]
                    local_reaction = \
                        self.results[case_name].node_reactions[uid][step]
                    global_reaction: nparr = np.array([
                        local_reaction[0],
                        local_reaction[1],
                        local_reaction[2],
                        local_reaction[3] + local_reaction[2] * y_coord
                        - local_reaction[1] * z_coord,
                        local_reaction[4] + local_reaction[0] * z_coord
                        - local_reaction[2] * x_coord,
                        local_reaction[5] + local_reaction[1] * x_coord
                        - local_reaction[0] * y_coord
                    ])
                    reactions += global_reaction
        return reactions


@dataclass
class StaticAnalysis(Analysis):
    """
    Static analysis.  Stores all results (for each load case) in one
    single step.
    """
    def run(self):
        """
        Runs the static analysis.
        """
        self._init_results()
        for case_name in self.load_cases:
            self._to_opensees_domain(case_name)
            self._define_loads(case_name)
            nodes = self.mdl.list_of_all_nodes()
            nodes.extend(self.load_cases[case_name].parent_nodes.values())
            elastic_elems = [
                elm for elm in self.mdl.list_of_elastic_beamcolumn_elements()
                if not elm.visibility.skip_opensees_definition]
            disp_elems = [
                elm for elm in self.mdl.list_of_disp_beamcolumn_elements()
                if not elm.visibility.skip_opensees_definition]
            zerolength_elems = self.mdl.list_of_zerolength_elements()
            step = 0
            ops.system(LN_ANALYSIS_SYSTEM)
            ops.numberer(NUMBERER)
            ops.constraints(*CONSTRAINTS)
            ops.test('NormDispIncr', 1.0e-8, 20, 3)
            ops.algorithm('Newton')
            ops.integrator('LoadControl', 1.0)
            ops.analysis('Static')
            ops.analyze(1)
            self._read_opensees_results(
                case_name,
                step,
                nodes,
                elastic_elems,
                disp_elems,
                zerolength_elems)
            if self.settings.pickle_results:
                self._write_results_to_disk()


@dataclass
class ModalAnalysis(Analysis):
    """
    Runs a modal analysis.
    """
    num_modes: int = field(default=1, repr=False)

    ####################################################
    # Methods that read back information from OpenSees #
    ####################################################

    def _read_node_displacements_modal(self, case_name):
        nodes = self.mdl.list_of_all_nodes()
        nodes.extend(self.load_cases[case_name].parent_nodes.values())
        for node in nodes:
            for i in range(self.num_modes):
                val = ops.nodeEigenvector(
                        node.uid,
                        i+1)
                self.results[case_name].node_displacements[node.uid][i] = val

    def _read_frame_element_forces_modal(self, case_name, elems):
        # note: opensees does not output the element forces that correspond
        # to the displacement field of each obtained mode.
        # to overcome this, we run a separate analysis imposing those
        # displacements to get the element forces.

        # note: work in progress. currently this doesn't fully work

        for case_name in self.load_cases:
            for step in range(self.num_modes):
                for elm in elems:
                    # if step == 3:
                    #     import pdb
                    #     pdb.set_trace()
                    assert isinstance(elm, ElasticBeamColumn)
                    # displacements at the two ends (global system)
                    u_i = (self.results[case_name].node_displacements
                           [elm.nodes[0].uid][step][0:3])
                    r_i = (self.results[case_name].node_displacements
                           [elm.nodes[0].uid][step][3:6])
                    u_j = (self.results[case_name].node_displacements
                           [elm.nodes[1].uid][step][0:3])
                    r_j = (self.results[case_name].node_displacements
                           [elm.nodes[1].uid][step][3:6])
                    offset_i = elm.geomtransf.offset_i
                    offset_j = elm.geomtransf.offset_j
                    u_i_o = transformations.offset_transformation(
                        offset_i, np.array(u_i), np.array(r_i))
                    u_j_o = transformations.offset_transformation(
                        offset_j, np.array(u_j), np.array(r_j))

                    x_vec = elm.geomtransf.x_axis
                    y_vec = elm.geomtransf.y_axis
                    z_vec: nparr = np.cross(x_vec, y_vec)

                    # global -> local transformation matrix
                    transf_global2local = \
                        transformations.transformation_matrix(
                            x_vec, y_vec, z_vec)
                    u_i_local = transf_global2local @ u_i_o
                    r_i_local = transf_global2local @ r_i
                    u_j_local = transf_global2local @ u_j_o
                    r_j_local = transf_global2local @ r_j

                    # # element UDL
                    udl = (self.load_cases[case_name]
                           .line_element_udl[elm.uid].val)
                    # note: modal analsis doesn't account for applied loads.
                    # this will cause issues with plotting if loads
                    # have been applied.
                    if np.linalg.norm(udl) > common.EPSILON:
                        raise ValueError('Loads applied at modal load case.')

                    # stiffness matrix terms
                    length = elm.clear_length()
                    etimesa = elm.section.e_mod * elm.section.area
                    etimesi_maj = elm.section.e_mod * elm.section.i_x
                    # eimin = elm.section.e_mod * elm.section.Iy
                    gtimesj = elm.section.g_mod * elm.section.j_mod

                    # deformations
                    d_l = u_j_local[0] - u_i_local[0]
                    theta_tor = r_j_local[0] - r_i_local[0]
                    du_xy = u_j_local[1] - u_i_local[1]
                    # du_xz = u_j_local[2] - u_i_local[2]

                    # axial load
                    n_i = etimesa / length * d_l

                    # strong axis bending
                    mz_i = \
                        (4.00 * etimesi_maj / length * r_i_local[2]
                         + 2.00 * etimesi_maj / length * r_j_local[2]
                         - 6.00 * etimesi_maj / length**2 * du_xy)
                    qy_i = (6.00 * etimesi_maj / length**2
                            * (r_i_local[2] + r_j_local[2])
                            - 12.00 * etimesi_maj / length**3 * du_xy)

                    # # weak axis bending
                    # # work in progress...
                    # myi = \
                    #     -(4.00 * eimin / length * (-r_i_local[1])
                    #      + 2.00 * eimin / length * (-r_j_local[1])
                    #      - 6.00 * eimin / length * (du_xz))
                    # qzi = (6.00 * eimin / length**2
                    #        * ((-r_i_local[2]) + (-r_j_local[2]))
                    #        - 12.00 * eimin / length**3 * (du_xz))
                    myi = 0.00
                    qzi = 0.00

                    # torsion
                    t_i = gtimesj / length * theta_tor

                    # store results
                    (self.results[case_name].element_forces
                     [elm.uid][step]) = \
                        np.array((n_i, qy_i, qzi, t_i, myi, mz_i))

    def run(self):
        """
        Runs the modal analysis.
        """
        self._init_results()
        for case_name in self.load_cases:
            self._to_opensees_domain(case_name)
            # tags = ops.getNodeTags()
            # self.print(len(tags))
            ops.constraints(*CONSTRAINTS)
            ops.system(MD_ANALYSIS_SYSTEM)
            # note: using SparseSYM results in wrong eigendecomposition
            num_inertial_nodes = 0
            ndtags = ops.getNodeTags()
            for node in ndtags:
                for j in range(6):
                    if ops.nodeMass(node, j+1) > 0.00:
                        num_inertial_nodes += 1
            eigenvalues: nparr = np.array(ops.eigen(
                self.num_modes))
            self.results[case_name].periods = 2.00*np.pi / np.sqrt(eigenvalues)
            self._read_node_displacements_modal(case_name)
            if self.settings.store_forces:
                self._read_frame_element_forces_modal(
                    case_name, self.mdl.list_of_beamcolumn_elements())
        if self.settings.pickle_results:
            self._write_results_to_disk()

    def modal_participation_factors(self, case_name, direction):
        """
        Calculates modal participation factors
        """
        dof_dir = {'x': 0, 'y': 1, 'z': 2}
        ntgs = ops.getNodeTags()
        gammas = np.zeros(self.num_modes)
        mstars = np.zeros(self.num_modes)
        mn_tot = 0.
        for ntg in ntgs:
            node_mass = self.load_cases[case_name].node_mass[ntg].val
            mn_tot += node_mass[dof_dir[direction]]
        for mode in range(self.num_modes):
            l_n = 0.
            m_n = 0.
            for ntg in ntgs:
                node_mass = self.load_cases[case_name].node_mass[ntg].val
                node_phi = ops.nodeEigenvector(ntg, mode+1)
                l_n += node_phi[dof_dir[direction]] * \
                    node_mass[dof_dir[direction]]
                for dof in range(6):
                    m_n += (node_phi[dof]**2) * node_mass[dof]
            gammas[mode] = l_n/m_n
            mstars[mode] = l_n**2/m_n
        mstars /= mn_tot
        return (gammas, mstars, mn_tot)


@dataclass
class NonlinearAnalysis(Analysis):
    """
    Nonlinear analysis parent class.
    """
    def _run_gravity_analysis(self, system):
        self.print(f'Setting system to {system}')
        ops.system(system)
        ops.numberer(NUMBERER)
        ops.constraints(*CONSTRAINTS)
        ops.test('NormDispIncr', 1.0e-6, 100, 3)
        ops.algorithm('RaphsonNewton')
        ops.integrator('LoadControl', 1)
        ops.analysis('Static')
        check = ops.analyze(1)
        if check != 0:
            raise ValueError('Analysis Failed')

    # def _acceptance_criteria(self):
    #     for elm in self.mdl.list_of_line_elements():

    #         if elm.model_as['type'] != 'elastic':
    #             continue
    #         mat = elm.section.material
    #         if mat.name == 'steel':
    #             capacity_t = mat.parameters['Fy']/mat.parameters['E0']
    #             capacity_c = -capacity_t
    #         else:
    #             raise ValueError('Unsupported material')
    #         strains = []
    #         x_vec = elm.x_axis
    #         y_vec = elm.y_axis
    #         z_vec = elm.z_axis
    #         trans_global2local: nparr = np.vstack((x_vec, y_vec, z_vec))
    #         forces_global: nparr = np.array(
    #             self.element_forces[str(elm.uid)][-1][0:3])
    #         moments_global_ends: nparr = np.array(
    #             self.element_forces[str(elm.uid)][-1][3:6])

    #         moments_global_clear = \
    #             transformations.offset_transformation(
    #                 elm.offset_i, moments_global_ends, forces_global)

    #         ni, qyi, qzi = trans_global2local @ forces_global
    #         ti, myi, mzi = trans_global2local @ moments_global_clear

    #         wx, wy, wz = elm.udl.total()

    #         len_clr = elm.length_clear
    #         t = np.linspace(0.00, len_clr, num=9)

    #         nx_vec = - t * wx - ni
    #         mz_vec = t**2 * 0.50 * wy + t * qyi - mzi
    #         my_vec = t**2 * 0.50 * wz + t * qzi + myi

    #         prop = elm.section.mesh.geometric_properties()
    #         area = prop['area']
    #         iy = prop['inertia']['ixx']
    #         iz = prop['inertia']['iyy']
    #         young_mod = elm.section.material.parameters['E0']

    #         for val in elm.section.snap_points.values():
    #             z, y = val
    #             stress = nx_vec/area \
    #                 + my_vec/iz * z \
    #                 - mz_vec/iy * y
    #             strain = stress / young_mod
    #             strains.extend(strain)
    #         emax = np.max(np.array(strains))
    #         emin = np.min(np.array(strains))
    #         if ((emax > capacity_t) or (emin < capacity_c)):
    #             raise ValueError(
    #                 "Acceptance criteria failed for element " +
    #                 str(elm.uid))

    def retrieve_node_displacement(self, uid, case_name):
        """
        Returns the displacement of a node for all analysis steps
        """
        res = np.zeros((self.results[case_name]
                        .n_steps_success, 6))  # type: ignore
        num = len(self.results[case_name].node_displacements[uid])
        for i in range(num):
            res[i] = self.results[case_name].node_displacements[uid][i]
        dframe = pd.DataFrame(res, columns=['ux', 'uy', 'uz',
                                            'urx', 'ury', 'urz'])
        dframe.index.name = 'step'
        return dframe

    def retrieve_node_acceleration(self, uid, case_name):
        """
        Returns the acceleration of a node for all analysis steps
        """
        res = np.zeros((self.results[case_name]
                        .n_steps_success, 6))  # type: ignore
        num = len(self.results[case_name].node_accelerations[uid])
        for i in range(num):
            res[i] = self.results[case_name].node_accelerations[uid][i]
        dframe = pd.DataFrame(res, columns=['ax', 'ay', 'az',
                                            'arx', 'ary', 'arz'])
        dframe.index.name = 'step'
        return dframe

    def retrieve_node_velocity(self, uid, case_name):
        """
        Returns the velocity of a node for all analysis steps
        """
        res = np.zeros((self.results[case_name].
                        n_steps_success, 6))  # type: ignore
        num = len(self.results[case_name].node_velocities[uid])
        for i in range(num):
            res[i] = self.results[case_name].node_velocities[uid][i]
        dframe = pd.DataFrame(res, columns=['vx', 'vy', 'vz',
                                            'vrx', 'vry', 'vrz'])
        dframe.index.name = 'step'
        return dframe

    def retrieve_node_abs_acceleration(self, uid, case_name):
        """
        Returns the absolute acceleration of a node for all analysis
        steps
        """
        res = np.zeros((self.results[case_name]
                        .n_steps_success, 6))  # type: ignore
        num = len(self.results[case_name].node_accelerations[uid])
        assert isinstance(self, NLTHAnalysis)
        assert self.a_g is not None
        for i in range(num):
            res[i] = self.results[case_name].node_accelerations[uid][i]
        for j in range(3):
            if j in self.a_g:
                a_g = interp1d(
                    self.a_g[j][:, 0], self.a_g[j][:, 1],
                    bounds_error=False, fill_value=0.00)
                res[:, j] += a_g(self.time_vector)*common.G_CONST_IMPERIAL
                # TODO: update to support SI
        dframe = pd.DataFrame(res, columns=['abs ax', 'abs ay', 'abs az',
                                            'abs arx', 'abs ary', 'abs arz'])
        dframe.index.name = 'step'
        return dframe

    def retrieve_node_abs_velocity(self, uid, case_name):
        """
        Returns the absolute velocity of a node for all analysis steps
        """
        res = np.zeros((self.results[case_name]
                        .n_steps_success, 6))  # type: ignore
        num = len(self.results[case_name].node_velocities[uid])
        assert isinstance(self, NLTHAnalysis)
        assert self.a_g is not None
        for i in range(num):
            res[i] = self.results[case_name].node_velocities[uid][i]
        for j in range(3):
            if j in self.a_g:
                a_g = interp1d(
                    self.a_g[j][:, 0], self.a_g[j][:, 1],
                    bounds_error=False, fill_value=0.00)
                v_g = integrate.cumulative_trapezoid(
                    a_g(self.time_vector)*common.G_CONST_IMPERIAL,
                    self.time_vector, initial=0)
                res[:, j] = res[:, j] + v_g
        dfrmae = pd.DataFrame(res, columns=['abs vx', 'abs vy', 'abs vz',
                                            'abs vrx', 'abs vry', 'abs vrz'])
        dfrmae.index.name = 'step'
        return dfrmae

    def retrieve_release_force_defo(self, uid, case_name):
        """
        Returns the force-deformation of a zerolength element for all
        analysis steps
        """
        num = len(self.results[case_name].release_force_defo[uid])
        res = np.zeros((num, 6))
        for i in range(num):
            res[i] = self.results[case_name].release_force_defo[uid][i]
        dframe = pd.DataFrame(res, columns=[
            'u1', 'u2', 'u3', 'q1', 'q2', 'q3'])
        dframe.index.name = 'step'
        return dframe


@dataclass
class PushoverAnalysis(NonlinearAnalysis):
    """
    Pushover analysis
    """
    def _apply_lateral_load(
            self, case_name, direction, modeshape=None, node=None):
        querry = LoadCaseQuerry(self.mdl, self.load_cases[case_name])
        distribution = querry.level_masses()
        distribution = distribution / np.linalg.norm(distribution)

        # define the load pattern
        ops.timeSeries('Linear', 2)
        ops.pattern("Plain", 2, 2)

        if direction == 'x':
            load_dir: nparr = np.array([1., 0., 0., 0., 0., 0.])
        elif direction == 'y':
            load_dir = np.array([0., 1., 0., 0., 0., 0.])
        elif direction == 'z':
            load_dir = np.array([0., 0., 1., 0., 0., 0.])
        else:
            raise ValueError('Invalid direction')

        if modeshape is not None:
            if direction not in ['x', 'y']:
                raise ValueError(
                    "Can't apply lateral loads based on the 1st " +
                    "mode shape in the z direction.")
            modeshape_ampl = modeshape / modeshape[-1]
        else:
            modeshape_ampl = np.ones(len(self.mdl.levels.values()))

        # if a node is given, apply the incremental load on that node
        if node:
            ops.load(node.uid, *(1.00*load_dir))
        else:
            for i, lvl in enumerate(self.mdl.levels.values()):
                if lvl.uid == 0:
                    continue
                if self.load_cases[case_name].parent_nodes:
                    node_list = [self.load_cases[case_name]
                                 .parent_nodes[lvl.uid]]
                else:
                    node_list = list(lvl.nodes.values())
                masses: nparr = np.array(
                    [self.load_cases[case_name].node_mass[
                        n.uid].val[0] for n in node_list])
                masses = masses/np.linalg.norm(masses)
                for j, some_node in enumerate(node_list):
                    ops.load(some_node.uid,
                             *(distribution[i]*masses[j]*load_dir *
                               modeshape_ampl[i]))

    def run(self, direction, target_displacements,
            control_node, displ_incr, modeshape=None, loaded_node=None):
        """
        Run pushover analysis
        Arguments:
          direction: can be any of 'x', 'y', 'z'
          target_displacements (list[float]): a list of target displcaments.
            each time a target is reached, the analysis continues until
            the next target is reached, flipping the direction as necessary.
          control_node (Node): analysis control node (of which the
            direciton is querried)
          displ_incr (float): initial displacement increment.
          modeshape (nparr): array containing a mode shape that is
            used to distribute the applied incremental loads. If no
            mode shape is specified, the distribution is uniform.
          loaded_node (Node): if a loaded node is specified, all
            incremental load is applied entirely on that node.
            Otherwise, the incremental loads are distributed to all nodes.
        """
        if direction == 'x':
            control_dof = 0
        elif direction == 'y':
            control_dof = 1
        elif direction == 'z':
            control_dof = 2
        else:
            raise ValueError("Direction can be 'x', 'y' or 'z'")

        self._init_results()

        for case_name in self.load_cases:
            nodes = self.mdl.list_of_all_nodes()
            nodes.extend(self.load_cases[case_name].parent_nodes.values())
            elastic_elems = [
                elm for elm in self.mdl.list_of_elastic_beamcolumn_elements()
                if not elm.visibility.skip_opensees_definition]
            disp_elems = [
                elm for elm in self.mdl.list_of_disp_beamcolumn_elements()
                if not elm.visibility.skip_opensees_definition]
            zerolength_elems = self.mdl.list_of_zerolength_elements()

            self._to_opensees_domain(case_name)
            self._define_loads(case_name)
            self._run_gravity_analysis(NL_ANALYSIS_SYSTEM)

            curr_displ = ops.nodeDisp(control_node.uid, control_dof+1)
            n_steps_success = 0
            self._read_opensees_results(
                case_name,
                n_steps_success,
                nodes,
                elastic_elems,
                disp_elems,
                zerolength_elems)

            ops.wipeAnalysis()
            ops.loadConst('-time', 0.0)
            self._apply_lateral_load(
                case_name, direction, modeshape, loaded_node)
            ops.numberer(NUMBERER)
            ops.constraints(*CONSTRAINTS)

            total_fail = False
            num_subdiv = 0
            num_times = 0

            scale = [1.0, 0.1, 0.01, 0.001, 0.0001, 0.00001]
            steps = [25, 50, 75, 100, 250, 500]
            norm = [1.0e-6] * 5 + [1.0e-2]

            try:

                for target_displacement in target_displacements:

                    if total_fail:
                        break

                    # determine push direction
                    if curr_displ < target_displacement:
                        displ_incr = abs(displ_incr)
                        sign = +1.00
                    else:
                        displ_incr = -abs(displ_incr)
                        sign = -1.00

                    self.print(f'entering loop: {target_displacement}')
                    while curr_displ * sign < target_displacement * sign:

                        # determine increment
                        if abs(curr_displ - target_displacement) < \
                           abs(displ_incr) * scale[num_subdiv]:
                            incr = sign * abs(curr_displ - target_displacement)
                        else:
                            incr = displ_incr * scale[num_subdiv]
                        ops.test('NormDispIncr', norm[num_subdiv],
                                 steps[num_subdiv], 0)
                        ops.algorithm('RaphsonNewton')
                        # ops.integrator("ArcLength", 1.00e1, 1.00e-7)
                        ops.integrator("DisplacementControl",
                                       int(control_node.uid), control_dof + 1,
                                       incr)
                        ops.system(NL_ANALYSIS_SYSTEM)
                        ops.analysis("Static")
                        flag = ops.analyze(1)
                        if flag != 0:
                            # analysis failed
                            if num_subdiv == len(scale) - 1:
                                # can't refine further
                                self.print('===========================')
                                self.print('Analysis failed to converge')
                                self.print('===========================')
                                total_fail = True
                                break
                            # can still reduce step size
                            num_subdiv += 1
                            # how many times to run with reduced step size
                            num_times = 10
                        else:
                            # analysis was successful
                            if num_times != 0:
                                num_times -= 1
                            n_steps_success += 1
                            self._read_opensees_results(
                                case_name,
                                n_steps_success,
                                nodes,
                                elastic_elems,
                                disp_elems,
                                zerolength_elems)
                            # self._acceptance_criteria()
                            curr_displ = ops.nodeDisp(
                                int(control_node.uid), control_dof+1)
                            print('Target displacement: '
                                  f'{target_displacement:.2f}'
                                  f' | Current: {curr_displ:.4f}', end='\r')
                            if num_subdiv != 0:
                                if num_times == 0:
                                    num_subdiv -= 1
                                    num_times = 10

            except KeyboardInterrupt:
                self.print("Analysis interrupted")

            n_steps_success += 1
            self._read_opensees_results(
                case_name,
                n_steps_success,
                nodes,
                elastic_elems,
                disp_elems,
                zerolength_elems)
            self.print(f'Number of saved analysis steps: {n_steps_success}')
            metadata = {'successful steps': n_steps_success}
            self.results[case_name].n_steps_success = n_steps_success
            self.results[case_name].metadata = metadata  # type: ignore
        # done with all cases.
        if self.settings.pickle_results:
            self._write_results_to_disk()

    def table_pushover_curve(self, case_name, direction, node):
        """
        Returns the force deformation results
        """
        if direction == 'x':
            control_dof = 0
        elif direction == 'y':
            control_dof = 1
        elif direction == 'z':
            control_dof = 2
        else:
            raise ValueError("Direction can be 'x', 'y' or 'z'")
        base_shear_lst = []
        displacement_lst = []
        for step in range(self.results[case_name]
                          .n_steps_success):  # type:ignore
            base_shear_lst.append(
                self.global_reactions(case_name, step)[control_dof])
            displacement_lst.append(
                self.results[case_name].node_displacements[
                    node.uid][step][control_dof])
        base_shear: nparr = -np.array(base_shear_lst)
        displacement: nparr = np.array(displacement_lst)
        return displacement, base_shear

    def plot_pushover_curve(self, case_name, direction, node):
        """
        Plots the pushover curve
        """
        # TODO: units
        displacement, base_shear = self.table_pushover_curve(
            case_name, direction, node)
        general_2d.line_plot_interactive(
            "Pushover Analysis Results<br>" + "Direction: " + direction,
            displacement, base_shear, 'spline+markers',
            "Displacement", "in", ".0f",
            "Base Shear", "lb", ".0f")

    # def plot_brace_hysteresis(self, brace):
    #     drift = []
    #     resisting_force = []
    #     n_i = brace.node_i.uid
    #     n_j = brace.node_j.uid
    #     x_axis = brace.x_axis
    #     x_axis_horiz: nparr = np.array((x_axis[0], x_axis[1], 0.00))
    #     x_axis_horiz = x_axis_horiz / np.linalg.norm(x_axis_horiz)
    #     for step in range(self.results[case_name].n_steps_success):
    #         disp_i = self.node_displacements[str(n_i)][step][0:3]
    #         disp_j = self.node_displacements[str(n_j)][step][0:3]
    #         diff_disp: nparr = np.array(disp_j) - np.array(disp_i)
    #         disp_prj = np.dot(diff_disp, x_axis)
    #         drift.append(disp_prj)
    #         ielm = brace.end_segment_i.internal_elems[-1].uid
    #         force = self.element_forces[str(ielm)][step][0:3]
    #         force_prj = - np.dot(force, x_axis_horiz)
    #         resisting_force.append(force_prj)
    #     general_2d.line_plot_interactive(
    #         "Brace Resisting Force",
    #         drift, resisting_force, 'line',
    #         'Story drift', 'in', '.0f',
    #         'Resisting Force', 'lb', '.0f')


def define_lateral_load_pattern(
        filename_x,
        filename_y,
        filename_z,
        file_time_incr):
    """
    Defines the load pattern for a time-history analysis from
    previously parsed files with a constant dt
    """

    error = True
    if filename_x:
        error = False
        # define X-direction TH
        ops.timeSeries('Path', 2, '-dt', file_time_incr,
                       '-filePath', filename_x,
                       '-factor', common.G_CONST_IMPERIAL)
        # pattern, direction, timeseries tag
        ops.pattern('UniformExcitation', 2, 1, '-accel', 2)

    if filename_y:
        error = False
        # define Y-direction TH
        ops.timeSeries('Path', 3, '-dt', file_time_incr,
                       '-filePath', filename_y,
                       '-factor', common.G_CONST_IMPERIAL)
        # pattern, direction, timeseries tag
        ops.pattern('UniformExcitation', 3, 2, '-accel', 3)

    if filename_z:
        error = False
        # define Z-direction TH
        ops.timeSeries('Path', 4, '-dt', file_time_incr,
                       '-filePath', filename_z,
                       '-factor', common.G_CONST_IMPERIAL)
        # pattern, direction, timeseries tag
        ops.pattern('UniformExcitation', 4, 3, '-accel', 4)

    if error:
        raise ValueError(
            "No input files specified.")


def plot_ground_motion(filename, file_time_incr,
                       gmunit='g',
                       plotly=False):
    """
    Plots a ground motion input file.
    """
    y_vals = np.loadtxt(filename)
    n_points = len(y_vals)
    x_vals = np.arange(0.00, n_points * file_time_incr, file_time_incr)
    if plotly:
        general_2d.line_plot_interactive(
            "Ground motion record<br>" +
            filename,
            x_vals, y_vals,
            "line",
            "Time", "s", ".3f",
            "Absolute Acceleration", gmunit, ".4f")
    else:
        plt.figure()
        plt.plot(x_vals, y_vals, 'k')
        plt.xlabel('Time (s)')
        plt.ylabel(f'Acceleration ({gmunit})')
        plt.show()


@dataclass
class NLTHAnalysis(NonlinearAnalysis):
    """
    Dynamic nonlinear time-history analysis
    """
    time_vector: list[float] = field(default_factory=list, repr=False)
    a_g: dict[int, npt.NDArray[np.float64]] = field(
        default_factory=dict, repr=False)

    def run(self,
            analysis_time_increment: float,
            filename_x: str,
            filename_y: str,
            filename_z: str,
            file_time_incr: float,
            finish_time=0.00,
            skip_steps=1,
            damping={'type': None},
            print_progress=True):
        """
        Run the nonlinear time-history analysis
        Args:
            filename_x, y, z: Paths where the fixed-step ground acceleration
                              records are stored (single-column).
            file_time_incr:   The corresponding time increment
            finish_time: Specify a target time (s) to stop the analysis
                         the default value of 0.00 means that it will
                         run for the entire duration of the files.
            damping: Can be any of:
                     {'type': None},
                     {'type': 'rayleigh', 'ratio': r, 'periods': [t1, t2]},
                     {'type': 'modal', 'num_modes': n}
            print_progress: Controls whether the current time is printed out
        """

        self._init_results()
        self.log('Running NLTH analysis')

        nodes = self.mdl.list_of_all_nodes()
        # note: only runs the first load case provided.
        # nlth should not have load cases.
        # will be fixed in the future.
        case_name = list(self.load_cases.keys())[0]
        nodes.extend(self.load_cases[case_name].parent_nodes.values())
        elastic_elems = [
            elm for elm in self.mdl.list_of_elastic_beamcolumn_elements()
            if not elm.visibility.skip_opensees_definition]
        disp_elems = [
            elm for elm in self.mdl.list_of_disp_beamcolumn_elements()
            if not elm.visibility.skip_opensees_definition]
        zerolength_elems = self.mdl.list_of_zerolength_elements()

        damping_type = damping.get('type')

        if damping_type == 'rayleigh':
            system = NL_ANALYSIS_SYSTEM
        elif damping_type == 'modal':
            system = MD_ANALYSIS_SYSTEM
        else:
            system = NL_ANALYSIS_SYSTEM

        nss = []
        if filename_x:
            gm_vals_x = np.genfromtxt(filename_x)
            nss.append(len(gm_vals_x))
        if filename_y:
            gm_vals_y = np.genfromtxt(filename_y)
            nss.append(len(gm_vals_y))
        if filename_z:
            gm_vals_z = np.genfromtxt(filename_z)
            nss.append(len(gm_vals_z))

        self.log(f'filename_x: {filename_x}')
        self.log(f'filename_y: {filename_y}')
        self.log(f'filename_z: {filename_z}')

        num_gm_points = np.min(np.array(nss))
        duration = num_gm_points * file_time_incr

        t_vec = np.linspace(0.00, file_time_incr*num_gm_points, num_gm_points)
        if filename_x:
            self.a_g[0] = np.column_stack((t_vec, gm_vals_x))
        else:
            self.a_g[0] = np.column_stack((t_vec, np.zeros(len(t_vec))))
        if filename_y:
            self.a_g[1] = np.column_stack((t_vec, gm_vals_y))
        else:
            self.a_g[1] = np.column_stack((t_vec, np.zeros(len(t_vec))))
        if filename_z:
            self.a_g[2] = np.column_stack((t_vec, gm_vals_z))
        else:
            self.a_g[2] = np.column_stack((t_vec, np.zeros(len(t_vec))))

        if finish_time == 0.00:
            target_timestamp = duration
        else:
            target_timestamp = finish_time

        self.log('Defining model in OpenSees')
        self._to_opensees_domain(case_name)

        # gravity analysis
        self.log('Defining dead loads')
        self._define_loads(case_name)
        self.log('Starting gravity analysis')
        self._run_gravity_analysis(system)
        n_steps_success = 0
        self._read_opensees_results(
            case_name,
            n_steps_success,
            nodes,
            elastic_elems,
            disp_elems,
            zerolength_elems
        )

        # time-history analysis
        ops.wipeAnalysis()
        ops.loadConst('-time', 0.0)
        curr_time = 0.00
        self.time_vector.append(curr_time)

        if damping_type == 'rayleigh':
            self.log('Using Rayleigh damping')
            w_i = 2 * np.pi / damping['periods'][0]
            zeta_i = damping['ratio']
            w_j = 2 * np.pi / damping['periods'][1]
            zeta_j = damping['ratio']
            a_mat: nparr = np.array([[1/w_i, w_i], [1/w_j, w_j]])
            b_vec: nparr = np.array([zeta_i, zeta_j])
            x_sol: nparr = np.linalg.solve(a_mat, 2*b_vec)
            ops.numberer(NUMBERER)
            ops.constraints(*CONSTRAINTS)
            ops.system(system)
            ops.rayleigh(x_sol[0], 0.0, 0.0, x_sol[1])
            # https://portwooddigital.com/2020/11/08/rayleigh-damping-coefficients/
            # thanks prof. Scott!

        if damping_type == 'modal':
            # tags = ops.getNodeTags()
            # num_nodes = len(tags) - 4
            # num_modeshapes = 3*num_nodes
            # self.print(len(tags))

            num_modes = damping['num_modes']
            # num_modes = num_modeshapes
            damping_ratio = damping['ratio']
            self.log(
                'Running eigenvalue analysis'
                f' with {num_modes} modes')
            ops.numberer(NUMBERER)
            ops.constraints(*CONSTRAINTS)
            ops.system(system)
            ops.eigen(num_modes)
            # ops.system(NL_ANALYSIS_SYSTEM)
            # ops.systemSize()
            self.log('Eigenvalue analysis finished')
            ops.modalDamping(damping['ratio'])
            self.log(
                f'{damping_ratio*100:.2f}% '
                'modal damping defined')

        self.log('Starting transient analysis')
        ops.test('NormDispIncr', 1e-6, 50, 0)
        ops.algorithm("KrylovNewton")
        # ops.integrator('Newmark', 0.50, 0.25)
        ops.integrator('TRBDF2')
        ops.analysis("Transient")

        define_lateral_load_pattern(
            filename_x,
            filename_y,
            filename_z,
            file_time_incr)

        num_subdiv = 0
        num_times = 0
        analysis_failed = False

        scale = [1.0, 0.1, 0.01, 0.001, 0.0001, 0.00001]

        total_step_count = 0
        now = time.perf_counter()

        try:

            while curr_time + common.EPSILON < target_timestamp:

                # # TODO create a method using this code
                # # to automate the fastest seup selection process
                # # debug
                # systems = ['SparseSYM', 'UmfPack']
                # algos = ['NewtonLineSearch',
                #          'ModifiedNewton',
                #          'KrylovNewton',
                #          'SecantNewton',
                #          'RaphsonNewton',
                #          'PeriodicNewton',
                #          'BFGS',
                #          'Broyden']
                # integrs = [('Newmark', 0.50, 0.25),
                #            ('HHT', 1.0),
                #            ('GeneralizedAlpha', 1., 1.),
                #            ('TRBDF2',)]

                # self.print('Testing systems')

                # times = []
                # best_sys = 'UmfPack'
                # best_algo = 'Newton'
                # best_integ = ('Newmark', 0.50, 0.25)
                # for syst in systems:
                #     ops.system(syst)
                #     ops.algorithm(best_algo)
                #     ops.integrator(*best_integ)

                #     tt = time.perf_counter()
                #     check = ops.analyze(
                #         20, analysis_time_increment * scale[num_subdiv])
                #     ttt = time.perf_counter()
                #     times.append(ttt - tt)
                # mn = min(times)
                # imn = times.index(mn)
                # best_sys = systems[imn]
                # self.print()
                # self.print(times)
                # self.print(best_sys)
                # self.print()

                # times = []
                # for algo in algos:
                #     ops.system(best_sys)
                #     ops.algorithm(algo)
                #     ops.integrator(*best_integ)
                #     tt = time.perf_counter()
                #     check = ops.analyze(
                #         20, analysis_time_increment * scale[num_subdiv])
                #     ttt = time.perf_counter()
                #     times.append(ttt - tt)
                # mn = min(times)
                # imn = times.index(mn)
                # best_algo = algos[imn]
                # self.print()
                # self.print(times)
                # self.print(best_algo)
                # self.print()
                # times = []
                # for integ in integrs:
                #     ops.system(best_sys)
                #     ops.algorithm(best_algo)
                #     ops.integrator(*integ)
                #     tt = time.perf_counter()
                #     check = ops.analyze(
                #         20, analysis_time_increment * scale[num_subdiv])
                #     ttt = time.perf_counter()
                #     times.append(ttt - tt)
                # mn = min(times)
                # imn = times.index(mn)
                # best_integ = integrs[imn]
                # self.print()
                # self.print(times)
                # self.print(best_integ)
                # self.print()

                check = ops.analyze(
                    1, analysis_time_increment * scale[num_subdiv])

                # analysis speed stats
                total_step_count += 1
                speed = total_step_count / (time.perf_counter() - now)
                if total_step_count % 50 == 0:
                    # provide run speed statistics
                    # debug
                    print(f'Average speed: {speed:.2f} steps/s')
                    self.log(f'Average speed: {speed:.2f} steps/s')
                if check != 0:
                    # analysis failed
                    if num_subdiv == len(scale) - 1:
                        self.print('===========================')
                        self.print('Analysis failed to converge')
                        self.print('===========================')
                        self.logger.warning(  # type: ignore
                            "Analysis failed"  # type: ignore
                            f" at time {curr_time:.5f}")  # type: ignore
                        analysis_failed = True
                        break
                    # can still reduce step size
                    num_subdiv += 1
                    # how many times to run with reduced step size
                    num_times = 10
                else:
                    # analysis was successful
                    curr_time = ops.getTime()
                    if num_times != 0:
                        num_times -= 1
                    if total_step_count % skip_steps == 0:
                        n_steps_success += 1
                        self._read_opensees_results(
                            case_name,
                            n_steps_success,
                            nodes,
                            elastic_elems,
                            disp_elems,
                            zerolength_elems
                        )
                        self.time_vector.append(curr_time)
                    if print_progress:
                        print('Target timestamp: '
                              f'{target_timestamp:.2f} s '
                              f'| Current: {curr_time:.4f} s',
                              end='\r')
                    if num_subdiv != 0:
                        if num_times == 0:
                            num_subdiv -= 1
                            num_times = 10

        except KeyboardInterrupt:
            self.print("Analysis interrupted")
            self.logger.warning("Analysis interrupted")  # type: ignore

        metadata = {'successful steps': n_steps_success,
                    'analysis_finished_successfully': not analysis_failed}
        self.results[case_name].n_steps_success = len(self.time_vector)
        self.log('Analysis finished')
        if self.settings.pickle_results:
            self._write_results_to_disk()
        return metadata

    def plot_node_displacement_history(self, case_name,
                                       node, direction,
                                       plotly=False):
        """
        Plots the displacement history of the specified node.
        """
        time_vec = self.time_vector
        uid = node.uid
        results = []
        for k in range(self.results[case_name].n_steps_success):  # type:ignore
            results.append(self.results[case_name].node_displacements[
                uid][k][direction])
        vals: nparr = np.array(results)
        if plotly:
            general_2d.line_plot_interactive(
                f"Node {uid} displacement history",
                time_vec, vals,
                "line",
                "Time", "s", ".3f",
                "Rel. Displacement", 'in', ".4f")
        else:
            plt.figure()
            plt.plot(time_vec, vals, 'k')
            plt.xlabel('Time (s)')
            plt.ylabel('Displacement (in)')
            plt.show()


@dataclass
class ModalResponseSpectrumAnalysis:
    """
    Modal response spectrum analysis
    """
    mdl: Model
    load_case: LoadCase
    num_modes: int
    periods: nparr
    spectrum: nparr
    direction: str
    modal_q0: Optional[nparr] = field(default=None)
    vb_modal: Optional[nparr] = field(default=None)
    anl: Optional[Analysis] = field(default=None)

    def run(self):
        """
        Run the modal response spectrum analysis
        """
        spectrum_ifun = interp1d(
            self.periods, self.spectrum, kind='linear')
        anl = ModalAnalysis(
            self.mdl, {self.load_case.name: self.load_case},
            num_modes=self.num_modes)
        anl.settings.pickle_results = False
        anl.settings.store_fiber = False
        anl.settings.store_forces = True
        anl.settings.store_reactions = False
        anl.run()
        case_name = self.load_case.name
        gammas, mstars, mtot = anl.modal_participation_factors(
            case_name, self.direction)
        periods = anl.results[case_name].periods
        if self.mdl.settings.imperial_units:
            g_const = common.G_CONST_IMPERIAL
        else:
            g_const = common.G_CONST_SI
        vb_modal = np.zeros(self.num_modes)
        modal_q0 = np.zeros(self.num_modes)
        for i in range(self.num_modes):
            vb_modal[i] = ((spectrum_ifun(periods[i]))  # type: ignore
                           * mstars[i]  # type: ignore
                           * mtot * g_const)  # type: ignore
            modal_q0[i] = (
                gammas[i] * (spectrum_ifun(periods[i])  # type: ignore
                             / (2.*np.pi / periods[i])**2  # type: ignore
                             * g_const))  # type: ignore
        self.modal_q0 = modal_q0
        self.vb_modal = vb_modal
        self.anl = anl

    def combined_node_disp(self, node_uid):
        """
        Returns the SRSS-combined node displacement of a node
        """
        all_vals = []
        assert self.anl is not None
        for i in range(self.num_modes):
            vals = (np.array(  # type: ignore
                self.anl.results[self.load_case.name]  # type: ignore
                .node_displacements[node_uid][i])  # type: ignore
                    * self.modal_q0[i])  # type: ignore
            all_vals.append(vals)
        all_vals_np: nparr = np.column_stack(all_vals)
        return np.sqrt(np.sum(all_vals_np**2, axis=1))

    def combined_node_disp_diff(self, node_i_uid, node_j_uid):
        """
        Returns the SRSS-combined displacement difference between two
        nodes
        """
        all_vals = []
        assert self.anl is not None
        for i in range(self.num_modes):
            vals_i = (np.array(  # type: ignore
                self.anl.results[self.load_case.name]  # type: ignore
                .node_displacements[node_i_uid][i])  # type: ignore
                      * self.modal_q0[i])  # type: ignore
            vals_j = (np.array(  # type: ignore
                self.anl.results[self.load_case.name]  # type: ignore
                .node_displacements[node_j_uid][i])  # type: ignore
                      * self.modal_q0[i])  # type: ignore
            vals = vals_i - vals_j
            all_vals.append(vals)
        all_vals_np: nparr = np.column_stack(all_vals)
        return np.sqrt(np.sum(all_vals_np**2, axis=1))

    def combined_basic_forces(self, element_uid):
        """
        Returns the SRSS-combined basic forces of a line element
        """
        all_vals = []
        assert self.anl is not None
        for i in range(self.num_modes):
            vals = (np.array(  # type: ignore
                self.anl.results[self.load_case.name]  # type: ignore
                .element_forces[element_uid][i])  # type: ignore
                    * self.modal_q0[i])  # type: ignore
            all_vals.append(vals)
        all_vals_np: nparr = np.column_stack(all_vals)
        return np.sqrt(np.sum(all_vals_np**2, axis=1))



"""
Common definitions used everywhere for consistency.
https://plotly.com/python/reference/
"""
#                          __
#   ____  ____ ___  ____ _/ /
#  / __ \/ __ `__ \/ __ `/ /
# / /_/ / / / / / / /_/ /_/
# \____/_/ /_/ /_/\__, (_)
#                /____/
#
# https://github.com/ioannis-vm/OpenSees_Model_Generator

# https://www.google.com/search?q=color+picker
GRID_COLOR = '#d1d1d1'
NODE_PRIMARY_COLOR = '#7ac4b7'
NODE_INTERNAL_COLOR = '#a832a8'
FRAME_COLOR = '#0f24db'
BEAM_MESH_COLOR = '#6d7aed'
FLOOR_COLOR = '#c4994f'
OFFSET_COLOR = '#a83256'
BISECTOR_COLOR = '#32a852'

if __name__ == "__main__":
    pass



"""
Model Generator for OpenSees ~ basic forces
"""

#                          __
#   ____  ____ ___  ____ _/ /
#  / __ \/ __ `__ \/ __ `/ /
# / /_/ / / / / / / /_/ /_/
# \____/_/ /_/ /_/\__, (_)
#                /____/
#
# https://github.com/ioannis-vm/OpenSees_Model_Generator

from __future__ import annotations
import numpy as np
import numpy.typing as npt
import pandas as pd
from ..solver import Analysis
from ..ops.element import ElasticBeamColumn
from ..ops.element import DispBeamColumn
from ..solver import ModalResponseSpectrumAnalysis

nparr = npt.NDArray[np.float64]

# pylint: disable=no-else-return


def basic_forces(
        anl: Analysis,
        case_name: str,
        step: int,
        elm: ElasticBeamColumn | DispBeamColumn,
        num_points: int,
        as_tuple: bool = False) -> object:
    """
    Returns the basic forces of a specified element
    """
    if isinstance(anl, ModalResponseSpectrumAnalysis):
        forces = anl.combined_basic_forces(elm.uid)
        w_x, w_y, w_z = (0.00, 0.00, 0.00)
    else:
        forces = anl.results[case_name].element_forces[
            elm.uid][step]
        w_x, w_y, w_z = anl.load_cases[case_name].line_element_udl[elm.uid].val

    n_i, qy_i, qz_i = forces[0:3]
    t_i, my_i, mz_i = forces[3:6]

    p_i = np.array(elm.nodes[0].coords) + elm.geomtransf.offset_i
    p_j = np.array(elm.nodes[1].coords) + elm.geomtransf.offset_j
    len_clr = np.linalg.norm(p_i - p_j)

    t_vec = np.linspace(0.00, len_clr, num=num_points)

    nx_vec = - t_vec * w_x - n_i
    qy_vec = t_vec * w_y + qy_i
    qz_vec = t_vec * w_z + qz_i
    tx_vec = np.full(num_points, -t_i)
    mz_vec = t_vec**2 * 0.50 * w_y + t_vec * qy_i - mz_i
    my_vec = t_vec**2 * 0.50 * w_z + t_vec * qz_i + my_i

    if as_tuple:
        return (nx_vec, qy_vec, qz_vec,
                tx_vec, mz_vec, my_vec)
    else:
        dframe = pd.DataFrame.from_dict(
            {'nx': nx_vec,
             'qy': qy_vec,
             'qz': qz_vec,
             'tx': tx_vec,
             'mz': mz_vec,
             'my': my_vec}
        )
        return dframe



"""
Model Generator for OpenSees ~ visibility
"""

#                          __
#   ____  ____ ___  ____ _/ /
#  / __ \/ __ `__ \/ __ `/ /
# / /_/ / / / / / / /_/ /_/
# \____/_/ /_/ /_/\__, (_)
#                /____/
#
# https://github.com/ioannis-vm/OpenSees_Model_Generator

from dataclasses import dataclass, field


@dataclass
class ElementVisibility:
    """
    Element visibility object.
    Controls whether an element is displayed in the plots
    and whether it is defined in OpenSees or not
    """
    hidden_when_extruded: bool = field(default=False)
    hidden_at_line_plots: bool = field(default=False)
    skip_opensees_definition: bool = field(default=False)


@dataclass
class NodeVisibility:
    """
    Node visibility object.
    Controls whether a node is displayed in the plots
    """
    connected_to_zerolength: bool = field(default=False)



"""
Model Generator for OpenSees ~ line
"""

#                          __
#   ____  ____ ___  ____ _/ /
#  / __ \/ __ `__ \/ __ `/ /
# / /_/ / / / / / / /_/ /_/
# \____/_/ /_/ /_/\__, (_)
#                /____/
#
# https://github.com/ioannis-vm/OpenSees_Model_Generator

from dataclasses import dataclass, field
from typing import Optional
import numpy as np
import numpy.typing as npt
from .import common

# pylint: disable=no-else-return

nparr = npt.NDArray[np.float64]


@dataclass
class Line:
    """
    Finite-length line segment object.
    Used internally whenever operations involving lines are reuired.
    Attributes:
      tag (str)
      start (numpy.ndarray): starting point
      end (numpy.ndarray): end point
    """
    tag: str
    start: nparr = field(repr=False)
    end: nparr = field(repr=False)

    def __repr__(self):
        res = ''
        res += 'Line object\n'
        res += f'  start: {self.start}\n'
        res += f'  end: {self.end}\n'
        return res

    def length(self):
        """
        returns the length of the line.
        """
        return np.linalg.norm(self.end - self.start)

    def direction(self):
        """
        returns a unit verctor pointing from the start to the end of
        the line.
        """
        return (self.end - self.start) / self.length()

    def intersect(self, other: 'Line'):
        """
        Calculates the intersection point of this line with another line.
        Returns None if the lines don't intersect.
        Note: 'line' is actually a finite-length line segment.
        Parameters:
          other (Line): the other line
        """
        ra_dir = self.direction()
        rb_dir = other.direction()
        mat: nparr = np.array(
            [
                [ra_dir[0], -rb_dir[0]],
                [ra_dir[1], -rb_dir[1]]
            ]
        )
        if np.abs(np.linalg.det(mat)) <= common.EPSILON:
            # The lines are parallel
            # in this case, we check if they have
            # a common starting or ending point
            # (we ignore the case of a common segment,
            #  as it has no practical use for our purposes).
            if np.linalg.norm(self.start - other.start) <= common.EPSILON:
                result = self.start
            elif np.linalg.norm(self.start - other.end) <= common.EPSILON:
                result = self.start
            elif np.linalg.norm(self.end - other.start) <= common.EPSILON:
                result = self.end
            elif np.linalg.norm(self.end - other.end) <= common. EPSILON:
                result = self.end
            else:
                result = None
        # Get the origins
        ra_ori = self.start
        rb_ori = other.start
        # System left-hand-side
        bvec: nparr = np.array(
            [
                [rb_ori[0] - ra_ori[0]],
                [rb_ori[1] - ra_ori[1]],
            ]
        )
        # Solve to get u and v in a vector
        uvvec = np.linalg.solve(mat, bvec)
        # Terminate if the intersection point
        # does not lie on both lines
        if uvvec[0] < 0 - common.EPSILON:
            result = None
        if uvvec[1] < 0 - common.EPSILON:
            result = None
        if uvvec[0] > self.length() + common.EPSILON:
            result = None
        if uvvec[1] > other.length() + common.EPSILON:
            result = None
        # Otherwise the point is valid
        point = ra_ori + ra_dir * uvvec[0]
        result = np.array([point[0], point[1]])

        return result

    def intersects_pt(self, point: nparr) -> bool:
        """
        Check whether the given point pt
        lies on the line
        Parameters:
            pt (nparr): a point
        """
        r_a = self.end - self.start
        norm2 = np.dot(r_a, r_a)
        if np.abs(norm2) < 1.0e-4:
            raise ValueError
        r_b = (point - self.start)
        cross = np.linalg.norm(np.cross(r_a, r_b))
        dot_normalized = np.dot(r_a, r_b)/norm2  # type: ignore
        if cross < common.EPSILON:
            res = bool(0.00 <= dot_normalized <= 1.00)
        else:
            res = False
        return res

    def point_distance(self, point: nparr) -> Optional[float]:
        """
        Calculates the projection of a point on the line.
        If the projection falls on the line segment, it returns the
        distance from the point to the line.
        Arguments:
          point (numpy.ndarray): the point's coordinates
        """
        r_a = self.end - self.start
        r_b = point - self.start
        proj_point = (r_b @ r_a) / (r_a @ r_a) * r_a
        if self.intersects_pt(proj_point + self.start):
            res: Optional[float] = float(np.linalg.norm(r_b - proj_point))
        else:
            res = None
        return res

    def project(self, point: nparr) -> Optional[nparr]:
        """
        Calculates the projection of a point on the line.
        If the projection falls on the line segment, it returns the
        projected point.
        Arguments:
          point (numpy.ndarray): the point's coordinates
        """
        r_a = self.end - self.start
        r_b = point - self.start
        proj_point = (r_b @ r_a) / (r_a @ r_a) * r_a + self.start
        if self.intersects_pt(proj_point):
            return proj_point
        else:
            return None



"""
Model Generator for OpenSees ~ load case
"""

#                          __
#   ____  ____ ___  ____ _/ /
#  / __ \/ __ `__ \/ __ `/ /
# / /_/ / / / / / / /_/ /_/
# \____/_/ /_/ /_/\__, (_)
#                /____/
#
# https://github.com/ioannis-vm/OpenSees_Model_Generator

# pylint: disable=W1512

from __future__ import annotations
from dataclasses import dataclass, field
from typing import TYPE_CHECKING
import numpy as np
import numpy.typing as npt
from . import transformations
from . import collections
from .preprocessing.tributary_area_analysis import TributaryAreaAnaysis
from .preprocessing.rigid_diaphragm import RDAnalyzer

if TYPE_CHECKING:
    from .model import Model
    from .ops.element import ElasticBeamColumn
    from .ops.node import Node

nparr = npt.NDArray[np.float64]


@dataclass
class PointLoadMass:
    """
    Point load/mass object. Global coordinate system.
    Attributes:
        other (list[float])
        floor (list[float])
    """
    val: nparr = field(
        default_factory=lambda: np.zeros(shape=6))

    def add(self, load: nparr):
        """
        Adds some quantity to the existing quantity
        """
        self.val += load

    def __repr__(self):
        res = ''
        res += 'Point Load (or mass) object\n'
        res += 'Components: (global system)\n'
        res += f'val: {self.val}\n'
        return res


@dataclass(repr=False)
class LineElementUDL:
    """
    Line element uniformly distributed load object.
    """
    parent_line_element: ElasticBeamColumn
    val: nparr = field(
        default_factory=lambda: np.zeros(shape=3))

    def __repr__(self):
        res = ''
        res += 'LineElementUDL object\n'
        res += f'parent_line_element.uid: {self.parent_line_element.uid}\n'
        res += 'Components:\n'
        res += f'  val: {self.val}\n'
        return res

    def add_glob(self, udl: nparr):
        """
        Adds a uniformly distributed load
        to the existing udl
        The load is defined
        with respect to the global coordinate system
        of the building, and it is converted to the
        local coordinate system prior to adding it.
        Args:
            udl (nparr): Array of size 3 containing
                components of the uniformly distributed load that is
                applied to the clear length of the element, acting on
                the global x, y, and z directions, in the direction of
                the global axes.
        """
        transf_mat = transformations.transformation_matrix(
            self.parent_line_element.geomtransf.x_axis,
            self.parent_line_element.geomtransf.y_axis,
            self.parent_line_element.geomtransf.z_axis)
        udl_local = transf_mat @ udl
        self.val += udl_local

    def to_global(self):
        """
        Returns the quantity expressed in the global coordinate system
        """
        udl = self.val
        transf_mat = transformations.transformation_matrix(
            self.parent_line_element.geomtransf.x_axis,
            self.parent_line_element.geomtransf.y_axis,
            self.parent_line_element.geomtransf.z_axis)
        return transf_mat.T @ udl


@dataclass(repr=False)
class LoadCase:
    """
    Load Case object.
    Load cases contain information related to the specified loads,
    mass, parent nodes and rigid diaphragm constraints, etc.
    Analysis objects can use multiple load cases.
    Load combination objects utilize load cases as well.
    """
    name: str
    parent_model: Model
    node_loads: collections.Collection[int, PointLoadMass] = field(init=False)
    node_mass: collections.Collection[int, PointLoadMass] = field(init=False)
    line_element_udl: collections.Collection[
        int, LineElementUDL] = field(init=False)
    tributary_area_analysis: collections.Collection[
        int, TributaryAreaAnaysis] = field(init=False)
    parent_nodes: dict[int, Node] = field(default_factory=dict)

    def __post_init__(self):
        self.node_loads = collections.Collection(self)
        self.node_mass = collections.Collection(self)
        self.line_element_udl = collections.Collection(self)
        self.tributary_area_analysis = \
            collections.Collection(self)
        # initialize loads and mass for each node and element
        for node in self.parent_model.list_of_all_nodes():
            self.node_loads[node.uid] = PointLoadMass()
            self.node_mass[node.uid] = PointLoadMass()
        for line_element in (self.parent_model
                             .list_of_beamcolumn_elements()):
            self.line_element_udl[line_element.uid] = \
                LineElementUDL(line_element)
        # initialize tributary area analysis for each level
        for lvlkey, lvl in self.parent_model.levels.items():
            self.tributary_area_analysis[lvlkey] = \
                TributaryAreaAnaysis(self, lvl)

    def rigid_diaphragms(self, level_uids: list[int], gather_mass=False):
        """
        Processes the geometry of the given levels and applies rigid
        diaphragm constraints
        """
        for lvl_uid in level_uids:
            lvl = self.parent_model.levels[lvl_uid]
            rda = RDAnalyzer(self, lvl)
            rda.run(gather_mass)

    def __repr__(self):
        res = ''
        res += 'LoadCase object\n'
        res += f'name: {self.name}\n'
        res += f'parent_model: {self.parent_model.name}\n'
        return res



"""
Model Generator for OpenSees ~ generic
"""

#                          __
#   ____  ____ ___  ____ _/ /
#  / __ \/ __ `__ \/ __ `/ /
# / /_/ / / / / / / /_/ /_/
# \____/_/ /_/ /_/\__, (_)
#                /____/
#
# https://github.com/ioannis-vm/OpenSees_Model_Generator


# pylint: disable=inconsistent-return-statements

from __future__ import annotations
from typing import TYPE_CHECKING
from typing import Union
from typing import Optional
from dataclasses import dataclass
import numpy as np
import numpy.typing as npt
from ..line import Line
from ..load_case import LoadCase
from .. import common
if TYPE_CHECKING:
    from ..component_assembly import ComponentAssembly
    from ..ops.element import ElasticBeamColumn
    from ..ops.element import DispBeamColumn
    from ..ops.node import Node
    from ..model import Model


nparr = npt.NDArray[np.float64]


@dataclass(repr=False)
class ElmQuerry:
    """
    Used by all component generators
    """
    model: Model

    def search_connectivity(
            self,
            nodes: list[Node]) -> Optional[ComponentAssembly]:
        """
        find component assembly based on connectivity
        """
        res = None
        uids = [node.uid for node in nodes]
        uids.sort()
        uids_tuple = (*uids,)
        conn_dict = self.model.component_connectivity()
        conn_dict.get(uids_tuple)
        if uids_tuple:
            res = conn_dict[uids_tuple]
        return res

    def search_node_lvl(self,
                        x_loc: float, y_loc: float,
                        lvl: int,
                        z_loc: Optional[float] = None,
                        internal=False
                        ) -> Optional[Node]:
        """
        Looks if a node exists at the given location.
        """
        lvls = self.model.levels
        level = lvls[lvl]
        res = None
        # check to see if node exists
        if z_loc:
            candidate_pt: nparr = np.array([x_loc, y_loc, z_loc])
            ndims = 3
        else:
            candidate_pt = np.array(
                [x_loc, y_loc])
            ndims = 2
        nodes = level.nodes
        if internal:
            for comp in level.components.values():
                nodes.update(comp.internal_nodes)
        for other_node in nodes.values():
            other_pt: nparr = np.array(other_node.coords[:ndims])
            if np.linalg.norm(candidate_pt - other_pt) < common.EPSILON:
                res = other_node
                break
        return res

    def retrieve_components_from_nodes(
            self,
            nodes: list[Node],
            lvl_uid: Optional[int] = None) -> dict[int, ComponentAssembly]:
        """
        Retrieves component assemblies if at least one of their
        external nodes matches the given list of nodes.
        """
        retrieved_components = {}
        if lvl_uid:
            level = self.model.levels[lvl_uid]
            candidate_components = level.components.values()
        else:
            candidate_components = self.model.list_of_components()
        given_node_uids = [n.uid for n in nodes]
        for component in candidate_components:
            accept = False
            external_nodes = component.external_nodes.values()
            for node in external_nodes:
                if node.uid in given_node_uids:
                    accept = True
                    continue
            if accept:
                retrieved_components[component.uid] = component
        return retrieved_components

    def retrieve_component(self, x_loc, y_loc, lvl):
        """
        Retrieves a component assembly of a level if any of its
        beamcolumn elements passes trhough the specified point.
        Returns the first element found.
        """
        level = self.model.levels[lvl]
        for component in level.components.values():
            if len(component.external_nodes) != 2:
                continue
            line_elems: list[Union[ElasticBeamColumn, DispBeamColumn]] = []
            line_elems.extend(component.elastic_beamcolumn_elements
                              .values())
            line_elems.extend(component.disp_beamcolumn_elements
                              .values())
            for elm in line_elems:
                p_i = (np.array(elm.nodes[0].coords)
                       + elm.geomtransf.offset_i)
                p_j = (np.array(elm.nodes[1].coords)
                       + elm.geomtransf.offset_j)
                if np.linalg.norm(p_i[0:2] - p_j[0:2]) < common.EPSILON:
                    if np.linalg.norm(
                            np.array((x_loc, y_loc))
                            - p_i[0:2]) < common.EPSILON:
                        return component
                else:
                    line = Line('', p_i[0:2], p_j[0:2])
                    line.intersects_pt(np.array((x_loc, y_loc)))
                    if line.intersects_pt(np.array((x_loc, y_loc))):
                        return component


@dataclass
class LoadCaseQuerry:
    """
    Load case querry object.
    """
    model: Model
    loadcase: LoadCase

    def level_masses(self):
        """
        Returns the total mass of each level.
        """
        mdl = self.model
        num_lvls = len(mdl.levels)
        distr = np.zeros(num_lvls)
        for key, lvl in mdl.levels.items():
            for node in lvl.nodes.values():
                mass = self.loadcase.node_mass[node.uid]
                distr[key] += mass.val[0]

            for component in lvl.components.values():
                for node in component.internal_nodes.values():
                    mass = self.loadcase.node_mass[node.uid]
                    distr[key] += mass.val[0]
        for uid, node in self.loadcase.parent_nodes.items():
            distr[uid] += self.loadcase.node_mass[node.uid].val[0]
        return distr
